# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from .. import logger
from ..authentication.auth import Auth
from ..custom_exceptions import EngineException
from ..user_config import EngineConfig
from . import EngineType


class EngineFactory:
    """
    Factory class of the Engine objects. It uses the server specific implementation depending on the value of
    ServerType.ETCD3
    """

    def __init__(self, engine_conf: EngineConfig, auth: Auth):
        """
        It uses the EngineConfig object to instantiate one instance of the server object. This instance contains the
        user's authentication details and the server URL. It is unique across the various Engine objects
        :param engine_conf:
        :param auth
        """
        assert engine_conf is not None, "Engine configuration required"
        assert engine_conf.host != "", "Server host is required"
        assert engine_conf.port != "", "Server port is required"
        assert engine_conf.type != "", "Server type is required"
        self._conf = engine_conf
        self._auth = auth

    def create_engine(self):
        """
        :return: an instance of the specific implementation of the Engine class
        """
        if self._conf.type == EngineType.ETCD_GRPC:
            # connect to the server by using pythonEtcd3
            logger.debug(f"Setting up gRPC stub to connect to the etcd server {self._conf.host}:{self._conf.port}")
        elif self._conf.type == EngineType.ETCD_REST:
            # connect to the server by using the REST API
            logger.debug(
                f"Setting up REST interface to connect to the etcd server " f"{self._conf.host}:{self._conf.port}"
            )
        elif self._conf.type == EngineType.FILE_BASED:
            # connect to the test file based server
            logger.debug("Setting up file-based test engine")
        else:
            raise EngineException(f"Configuration error - Engine: {self._conf.type} is not recognised")

        # instantiate the engine and return it
        engine_class = self._conf.type.get_class()
        try:
            # instantiate the engine and return
            return engine_class(config=self._conf, auth=self._auth)
        except Exception as e:
            raise EngineException(f"Error in creating the engine {engine_class.__name__}: {e}")



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import importlib
from enum import Enum


class AuthType(Enum):
    """
    This Enum describes the various authentication currently available.
    """

    ECMWF = ("ecmwf_auth", "EcmwfAuth")
    ETCD = ("etcd_auth", "EtcdAuth")
    NONE = ("none_auth", "NoneAuth")

    def __str__(self):
        return self.name.lower()

    def get_class(self):
        module = importlib.import_module("pyaviso.authentication." + self.value[0])
        return getattr(module, self.value[1])



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from abc import ABC, abstractmethod


class TypeHandler(ABC):
    def __init__(self, key, required=False):
        super(TypeHandler, self).__init__()
        self._key = key
        self._required = required

    @property
    def key(self) -> str:
        return self._key

    @property
    def required(self) -> bool:
        return self._required

    def process(self, value: any = None) -> str:
        if self.required and value is None:
            raise KeyError(f"{self.key} is a mandatory key")
        elif not self.required and value is None:
            return None
        else:
            # validate the key
            if self.valid(value):
                return self.canonise(value)
            else:
                raise ValueError(f"Value {value} is not valid for key {self.key}")

    @abstractmethod
    def valid(self, value: any) -> bool:
        pass

    @abstractmethod
    def canonise(self, value: any) -> str:
        pass



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import base64
import http.client
import logging
import time
from typing import Dict, List

import requests

from .. import logger
from ..authentication.auth import Auth
from ..authentication.etcd_auth import EtcdAuth
from ..custom_exceptions import EngineException, EngineHistoryNotAvailableError
from ..user_config import EngineConfig
from .etcd_engine import MAX_KV_RETURNED, EtcdEngine


class EtcdRestEngine(EtcdEngine):
    """
    This class is a specialisation of the Engine class, able to connect to a etcd3 server via the gRPC gateway by
    relying on the standard REST requests library.
    """

    def __init__(self, config: EngineConfig, auth: Auth):
        super(EtcdRestEngine, self).__init__(config, auth)
        # set base url
        if self.https:
            self._base_url = f"https://{self._host}:{self._port}/v3/"
        else:
            self._base_url = f"http://{self._host}:{self._port}/v3/"

    def pull(
        self,
        key: str,
        key_only: bool = False,
        rev: int = None,
        prefix: bool = True,
        min_rev: int = None,
        max_rev: int = None,
    ) -> List[Dict[str, any]]:
        """
        This method implements a query to the notification server for all the key-values associated to the key as input.
        This key by default is a prefix, it can therefore return a set of key-values
        :param key: input in the query
        :param key_only: if True no values are returned
        :param rev: revision to pull
        :param prefix: if true the function will retrieve all the KV pairs starting with the key passed
        :param min_rev: if provided it filters for only KV pairs with mod_revision >= to min_rev
        :param max_rev: if provided it filters for only KV pairs with mod_revision <= to max_rev
        :return: List of key-value pairs formatted as dictionary
        """
        logger.debug(f"Calling pull for {key}...")

        url = self._base_url + "kv/range"

        # determine the range_end
        if prefix:
            range_end = self._encode_to_str_base64(str(self._incr_last_byte(key), "utf-8"))
        else:
            range_end = None

        # first authenticate and use the token for the header
        self._authenticate()

        # encode key
        encoded_key = self._encode_to_str_base64(key)

        # create the body for the get range on the etcd sever, order them newest first
        body = {
            "key": encoded_key,
            "range_end": range_end,
            "limit": MAX_KV_RETURNED,
            "sort_order": "DESCEND",
            "sort_target": "KEY",
            "keys_only": key_only,
            "revision": rev,
            "min_mod_revision": min_rev,
            "max_mod_revision": max_rev,
        }
        # make the call
        logger.debug(f"Pull request: {body}")

        # start an infinite loop of request if the server side is unreachable
        while True:
            try:
                resp = requests.post(url, json=body, headers=self.auth.header(), timeout=self.timeout)
                resp.raise_for_status()
            except requests.exceptions.HTTPError as err:
                if resp.status_code == 408 or (resp.status_code >= 500 and resp.status_code < 600):
                    logger.warning(f"Unable to connect to {url}, trying again in {self.automatic_retry_delay}s...")
                    logger.debug(f"Not able to pull key {key}, {str(err)}, trying again...")
                    time.sleep(self.automatic_retry_delay)
                    continue
                elif resp.status_code == 400 and (
                    "History not available" in resp.content.decode()
                    or "required revision has been compacted" in resp.content.decode()
                ):
                    raise EngineHistoryNotAvailableError()
                else:
                    raise EngineException(f"Not able to pull key {key}, {str(err)}")
            except (requests.exceptions.ConnectionError, requests.exceptions.Timeout) as err:
                logger.warning(f"Unable to connect to {url}, trying again in {self.automatic_retry_delay}s...")
                logger.debug(f"Not able to pull key {key}, {str(err)}, trying again...")
                time.sleep(self.automatic_retry_delay)
                continue
            except Exception as e:
                logger.exception(e)
                raise EngineException(
                    f"Not able to pull key {key}, status {resp.status_code}, {resp.reason}, " f"{resp.content.decode()}"
                )

            # we got a good responce, exit from the loop
            break

        logger.debug(f"Query for {key} completed")

        # parse the result to return just key-value pairs
        new_kvs: List[Dict[str, bytes]] = []
        resp_body = resp.json()
        if "kvs" in resp_body:
            logger.debug("Building key-value list")
            for kv in resp_body["kvs"]:
                new_kv = self._parse_raw_kv(kv, key_only)
                new_kvs.append(new_kv)
                logger.debug(f"Key: {new_kv['key']} pulled successfully")

        logger.debug(f"{len(new_kvs)} keys found")
        return new_kvs

    def delete(self, key: str, prefix: bool = True) -> List[Dict[str, bytes]]:
        """
        This method deletes all the keys associated to this key, the key is a prefix as default
        :param key: key prefix to delete
        :param prefix: if true the function will delete all the KV pairs starting with the key passed
        :return: kvs deleted
        """
        logger.debug(f"Calling delete for {key}...")

        url = self._base_url + "kv/deleterange"

        # determine the range_end
        if prefix:
            range_end = self._encode_to_str_base64(str(self._incr_last_byte(key), "utf-8"))
        else:
            range_end = None

        # first authenticate and use the token for the header
        self._authenticate()

        # encode key
        encoded_key = self._encode_to_str_base64(key)

        # create the body for the delete range
        body = {"key": encoded_key, "range_end": range_end, "prev_kv": True}
        # make the call
        logger.debug(f"Deleting key range associated to key {key}")
        try:
            resp = requests.post(url, json=body, headers=self.auth.header(), timeout=self.timeout)
            resp.raise_for_status()
        except Exception as err:
            raise EngineException(f"Not able to delete key {key}, {str(err)}")

        logger.debug(f"Delete request for key {key} completed")

        # parse the result to return just key-value pairs of what has been deleted
        del_kvs: List[Dict[str, bytes]] = []
        resp_body = resp.json()
        if "prev_kvs" in resp_body:
            logger.debug("Building key-value list")
            for kv in resp_body["prev_kvs"]:
                new_kv = self._parse_raw_kv(kv)
                del_kvs.append(new_kv)
                logger.debug(f"Key: {new_kv['key']} deleted successfully")

        return del_kvs

    def push(self, kvs: List[Dict[str, any]], ks_delete: List[str] = None, ttl: int = None) -> bool:
        """
        Method to submit a list of key-value pairs and delete a list of keys from the server as a single transaction
        :param kvs: List of KV pair
        :param ks_delete: List of keys to delete before the push of the new ones. Note that each key is read as a folder
        :param ttl: time to leave of the keys pushed, once expired the keys will be deleted
        :return: True if successful
        """
        logger.debug("Calling push...")
        url = self._base_url + "kv/txn"

        # first authenticate and use the token for the header
        self._authenticate()

        # check if we need to request a lease for the ttl
        if ttl:
            lease = self._lease(ttl)

        logger.debug("Preparing the transaction statement")
        ops = []
        # first delete the keys requested
        if ks_delete is not None and len(ks_delete) != 0:
            for kd in ks_delete:
                # every key is deleted with prefix=True
                range_end = self._encode_to_str_base64(str(self._incr_last_byte(kd), "utf-8"))
                k = self._encode_to_str_base64(kd)
                delete = {"requestDeleteRange": {"key": k, "range_end": range_end}}
                ops.append(delete)

        # Prepare the transaction with a put operation for each KV pair
        for kv in kvs:
            k = self._encode_to_str_base64(kv["key"])
            v = self._encode_to_str_base64(kv["value"])
            put = {"requestPut": {"key": k, "value": v}}
            if ttl:
                put["requestPut"]["lease"] = lease
            ops.append(put)

        body = {"success": ops}
        # commit transaction
        # logger.debug(f"Committing the transaction statement: {body}")
        try:
            resp = requests.post(url, json=body, headers=self.auth.header(), timeout=self.timeout)
            resp.raise_for_status()
        except Exception as err:
            raise EngineException(f"Not able to execute the transaction, {str(err)}")

        logger.debug("Transaction completed")
        resp_body = resp.json()
        # read the header
        if "header" in resp_body:
            h = resp_body["header"]
            rev = int(h["revision"])
            logger.debug(f"New server revision {rev}")

        return True

    def _authenticate(self) -> bool:
        """
        This method authenticates  the user and set the internal token, this is only done for Etcd authentication
        :return: True if successfully authenticated
        """
        if type(self.auth) == EtcdAuth:
            logger.debug(f"Authenticating user {self.auth.username}...")

            url = self._base_url + "auth/authenticate"
            body = {"name": self.auth.username, "password": self.auth.password}
            try:
                resp = requests.post(url, json=body, headers=self.auth.header(), timeout=self.timeout)
                resp.raise_for_status()
            except Exception as err:
                raise EngineException(f"Not able to authenticate {self.auth.username}, {str(err)}")
            assert resp.json().get("token") is not None, "No token found in authentication response"
            self.auth.token = resp.json()["token"]

            logger.debug(f"User {self.auth.username} successfully authenticated")

        return True

    def _latest_revision(self, key: str) -> int:
        """
        :param: key used for the server request
        :return: latest revision of the notification server.
        """
        logger.debug("Querying notification server for latest revision")

        url = self._base_url + "kv/range"

        # first authenticate and use the token for the header
        self._authenticate()

        # we need just the header back from the server
        encoded_key = self._encode_to_str_base64(key)
        body = {"key": encoded_key, "keys_only": True}
        # make the call
        while True:
            try:
                resp = requests.post(url, json=body, headers=self.auth.header(), timeout=self.timeout)
                resp.raise_for_status()
            except requests.exceptions.HTTPError as err:
                if resp.status_code == 408 or (resp.status_code >= 500 and resp.status_code < 600):
                    logger.warning(f"Unable to connect to {url}, trying again in {self.automatic_retry_delay}s...")
                    logger.debug(f"Not able to request latest revision, {str(err)}, trying again...")
                    time.sleep(self.automatic_retry_delay)
                    continue
                else:
                    raise EngineException(f"Not able to request latest revision, {str(err)}")
            except (requests.exceptions.ConnectionError, requests.exceptions.Timeout) as err:
                logger.warning(f"Unable to connect to {url}, trying again in {self.automatic_retry_delay}s...")
                logger.debug(f"Not able to request latest revision, {str(err)}, trying again...")
                time.sleep(self.automatic_retry_delay)
                continue
            except Exception as e:
                logger.exception(e)
                raise EngineException(
                    f"Not able to request latest revision, status {resp.status_code}, {resp.reason}, "
                    f"{resp.content.decode()}"
                )

            # we got a good responce, exit from the loop
            break

        logger.debug("Query for latest revision completed")
        resp_body = resp.json()
        # read the header
        if "header" in resp_body:
            h = resp_body["header"]
            rev = int(h["revision"])
        else:
            raise EngineException("Error in reading server revision. Response does not contain header")
        logger.debug(f"Latest revision {rev}")
        return rev

    def _lease(self, ttl) -> str:
        """
        This method requests a Lease for the TTL specified
        :param ttl: Lease TTL
        :return: lease id
        """
        logger.debug(f"Calling lease for ttl {ttl}...")

        url = self._base_url + "lease/grant"

        # first authenticate and use the token for the header
        self._authenticate()

        # create the request body
        body = {"TTL": ttl, "ID": 0}

        # make the call
        try:
            resp = requests.post(url, json=body, headers=self.auth.header(), timeout=self.timeout)
            resp.raise_for_status()
        except Exception as err:
            raise EngineException(f"Not able to request a lease, {str(err)}")

        logger.debug("Lease request completed")
        resp_body = resp.json()
        if "ID" in resp_body:
            logger.debug(f"Lease {resp_body.get('ID')} acquired")
            return resp_body.get("ID")
        else:
            logger.error(f"Not able to read lease id from {resp_body}")
            raise EngineException("Not able to acquire lease")

    def _parse_raw_kv(self, kv: Dict[str, any], key_only: bool = False) -> Dict[str, any]:
        """
        Internal method to translate the kv pair coming from the etcd server into a dictionary that fits better this
        application
        :param kv: raw kv pair from the etcd server
        :param key_only:
        :return: translated kv pair as dictionary
        """
        new_kv = {}
        if not key_only:
            new_kv["value"] = self._decode_to_bytes(kv["value"])  # leave it as binary
        new_kv["key"] = self._decode_to_bytes(kv["key"]).decode()
        new_kv["version"] = int(kv["version"])
        new_kv["create_rev"] = int(kv["create_revision"])
        new_kv["mod_rev"] = int(kv["mod_revision"])
        return new_kv

    def _encode_to_str_base64(self, obj: any) -> str:
        """
        Internal method to translate the object passed in a field that could be accepted by etcd and the request library
        for the key or value. The request library accepts only strings encoded in base64 while etcd wants binaries for
        the key and value fields.
        :param obj:
        :return: a base64 string representation of the binary translation
        """
        if type(obj) is bytes:
            binary = obj
        elif type(obj) is str:
            binary = obj.encode()
        else:
            binary = str(obj).encode()

        return str(base64.b64encode(binary), "utf-8")

    def _decode_to_bytes(self, string: str) -> any:
        """
        Internal method to translate what is coming back from the notification server.
        The request library returns only string base64 encoded
        :param string:
        :return: the payload decoded from the base64 string representation
        """
        return base64.decodebytes(string.encode())


# Enable HTTPConnection debug logging to the logging framework
httpclient_logger = logging.getLogger("http.client")


def httpclient_log(*args):
    to_be_logged = " ".join(args)
    if len(to_be_logged) > 1000:
        to_be_logged = to_be_logged[:1000]
    httpclient_logger.log(logging.DEBUG, to_be_logged)


# mask the print() built-in in the http.client module to use logging instead
http.client.print = httpclient_log
http.client.HTTPConnection.debuglevel = 1



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from queue import Queue

from .version import __version__  # noqa: F401

# setting application logger - this goes before anything else
logger = logging.getLogger("aviso")
logger.setLevel(logging.DEBUG)

# home folder for configuration, state and log
HOME_FOLDER = "~/.aviso"

# system folder for configuration
SYSTEM_FOLDER = "/etc/aviso"

# This is a thread-safe communication channel. It is used to tell the main thread when to terminate.
exit_channel = Queue()

from .notification_manager import NotificationManager  # noqa: F401, E402



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from typing import Dict

from .. import logger
from . import trigger
from .trigger import TriggerType


class LogTrigger(trigger.Trigger):
    """
    This class implements the 'Log' trigger by logging to the log file specified the
    notification received
    """

    def __init__(self, notification: Dict[str, any], params: Dict[str, any]):
        trigger.Trigger.__init__(self, notification, params)
        self.trigger_type = TriggerType.log

    def execute(self):
        logger.info("Starting Log Trigger...")
        # create a file handler for the log specified
        log_path = self.params.get("path")
        handler = logging.FileHandler(log_path, "a")
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        handler.setFormatter(formatter)
        # get the logger and the new file handler to it
        log = logging.getLogger()
        log.addHandler(handler)
        # log the notification
        logger.info(f"Notification received: {self.notification}")
        # remove the logger
        log.removeHandler(handler)
        logger.info("Log Trigger completed")



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import functools
import signal
import sys
import threading
import time
from typing import Dict, List

import click

from pyaviso import __version__, logger
from pyaviso import user_config as conf
from pyaviso.custom_exceptions import (
    EngineException,
    EventListenerException,
    InvalidInputError,
    TriggerException,
)
from pyaviso.engine import EngineType
from pyaviso.notification_manager import NotificationManager
from pyaviso.service_config_manager import ServiceConfigException

# Create the listener manager
manager: NotificationManager = NotificationManager()

# set of known exceptions
KNOWN_EXCEPTION = (
    ServiceConfigException,
    EventListenerException,
    TriggerException,
    EngineException,
    InvalidInputError,
    AssertionError,
    KeyError,
)


def catch_all_exceptions(cls, handler):
    """
    This function is used to pass a child of the click.command class to the click CLI initialisation.
    This new class overrides the default error handling by allowing to intercept keyboard interruption and EOF errors.
    :param cls: click.command
    :param handler: function in charge of error handling
    :return:
    """

    class Cls(cls):

        _original_args = None

        def make_context(self, info_name, args, parent=None, **extra):

            # grab the original command line arguments
            self._original_args = " ".join(args)

            try:
                return super(Cls, self).make_context(info_name, args, parent=parent, **extra)
            except Exception:
                # call the handler
                handler()

                # let the user see the original error
                raise

        def invoke(self, ctx):
            try:
                return super(Cls, self).invoke(ctx)
            except Exception:
                # call the handler
                handler()

                # let the user see the original error
                raise

    return Cls


def ignore_signal(signum, frame):
    """
    This is used to ignore a specific signal sent to this process
    :param signum:
    :param frame:
    :return:
    """
    pass


def ignore_signal_and_sleep(signum, frame, time_sec=0.1):
    """
    This is used to ignore and sleep when a specific signal is sent to this process.
    The sleep is required when the signal is sent multiple times like the SIGTTIN in case of CLICK running in the
    background and trying to read from the stdin.
    :param time_sec: time in second to sleep
    :param signum:
    :param frame:
    :return:
    """
    time.sleep(time_sec)


def stop_listeners(signum=None, frame=None):
    """
    This function takes care of gracefully stopping the listeners.
    :param signum:
    :param frame:
    :return:
    """
    # Stop gracefully the notification listeners
    try:
        logger.debug("Stopping listeners...")
        manager.listener_manager.cancel_listeners()
        logger.info("Listeners stopped")
    except Exception as e:
        logger.error(f"Error while stopping the listeners, {e}")
        logger.debug("", exc_info=True)
        sys.exit(-1)


def stop_listeners_and_exit(signum=None, frame=None):
    """
    This function takes care of gracefully stopping the listeners and then exit
    propagates the exception.
    :param signum:
    :param frame:
    :return:
    """
    # Stop gracefully the notification listeners and exit
    stop_listeners()
    sys.exit()


def notification_server_setup(f):
    @click.option("--host", "-H", help="Notification server host.")
    @click.option("--port", "-P", help="Notification server port.", type=int)
    @click.option("--test", help="Activate TestMode.", is_flag=True, default=False)
    @functools.wraps(f)
    def functor(*args, **kwargs):
        if kwargs["host"]:
            kwargs["configuration"].notification_engine.host = kwargs["host"]
        kwargs.pop("host")
        if kwargs["port"]:
            kwargs["configuration"].notification_engine.port = kwargs["port"]
        kwargs.pop("port")
        if kwargs["test"]:
            kwargs["configuration"].notification_engine.type = EngineType.FILE_BASED
        kwargs.pop("test")

        return f(*args, **kwargs)

    return functor


def user_config_setup(f):
    @click.option("--config", "-c", help="User configuration file path.")
    @click.option("--log", "-l", help="Logging configuration file path.")
    @click.option("--debug", "-d", help="Enable the debug log.", is_flag=True, default=False)
    @click.option(
        "--quiet", "-q", help="Suppress non-error messages from the console output.", is_flag=True, default=False
    )
    @click.option("--no-fail", help="Suppress any error exit code.", is_flag=True, default=False)
    @click.option("--username", "-u", help="Username required to authenticate to the server.")
    @click.option("--key", "-k", help="File path to the key required to authenticate to the server.")
    @functools.wraps(f)
    def functor(*args, **kwargs):
        # CLIK automatically sets the flags, put back None values like for the other parameters
        kwargs["debug"] = None if not kwargs["debug"] else True
        kwargs["quiet"] = None if not kwargs["quiet"] else True
        kwargs["no_fail"] = None if not kwargs["no_fail"] else True

        # create the configuration object
        configuration = conf.UserConfig(
            conf_path=kwargs["config"],
            logging_path=kwargs["log"],
            debug=kwargs["debug"],
            quiet=kwargs["quiet"],
            no_fail=kwargs["no_fail"],
            username=kwargs["username"],
            key_file=kwargs["key"],
        )

        # pass it as a option in the same dictionary but remove the fields used for the configuration
        kwargs["configuration"] = configuration
        kwargs.pop("config")
        kwargs.pop("log")
        kwargs.pop("debug")
        kwargs.pop("quiet")
        kwargs.pop("no_fail")
        kwargs.pop("username")
        kwargs.pop("key")

        logger.debug(f"Running Aviso v.{__version__}")
        logger.debug(f"Configuration loaded: {configuration}")

        return f(*args, **kwargs)

    return functor


CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(version=__version__)
def cli():
    pass


@click.command(cls=catch_all_exceptions(click.Command, handler=stop_listeners), context_settings=CONTEXT_SETTINGS)
@user_config_setup
@notification_server_setup
@click.argument("listener_files", nargs=-1)
@click.option(
    "--from",
    "from_date",
    type=click.DateTime(formats=["%Y-%m-%dT%H:%M:%S.%fZ"]),
    help="Replay notification from this date.",
)
@click.option(
    "--to", "to_date", type=click.DateTime(formats=["%Y-%m-%dT%H:%M:%S.%fZ"]), help="Replay notification to this date."
)
@click.option("--now", "now", is_flag=True, default=False, help="Ignore missed notifications, only listen to new ones.")
@click.option("--catchup", "catchup", is_flag=True, default=False, help="Retrieve first the missed notifications.")
def listen(listener_files: List[str], configuration: conf.UserConfig, from_date, to_date, now, catchup):
    """
    This method allows the user to execute the listeners defined in the YAML listener file

    :param listener_files: YAML files used to define the listeners
    """
    try:
        """
        UNIX Signal handling
        """

        if threading.current_thread() is threading.main_thread():
            # This is needed to avoid the process to be suspended in case it runs in background, we must sleep
            # because we constantly read from the stdin
            signal.signal(signal.SIGTTIN, ignore_signal_and_sleep)
            # this is sent with CTRL + \
            signal.signal(signal.SIGQUIT, stop_listeners_and_exit)
            # this is sent whit the default kill command
            signal.signal(signal.SIGTERM, stop_listeners_and_exit)

        # call the main listen method
        manager.listen(
            configuration,
            listeners_file_paths=listener_files,
            from_date=from_date,
            to_date=to_date,
            now=now,
            catchup=catchup,
        )

    except KNOWN_EXCEPTION as e:
        logger.error(f"{e}")
        logger.debug("", exc_info=True)
        stop_listeners()
        sys.exit(-1)
    except Exception as e:
        logger.error(f"Error occurred while running the listeners: {e}")
        logger.debug("", exc_info=True)
        stop_listeners()
        sys.exit(-1)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("parameters", required=True)
@user_config_setup
@notification_server_setup
def key(parameters: str, configuration: conf.UserConfig):
    """
    Generate the key to send to the notification server according to the current schema using the parameters defined

    :param parameters: key1=value1,key2=value2,...
    """

    try:
        parsed_param = _parse_inline_params(parameters)
        # base_key and maintenance are ignored because not needed here
        key_generated, base_key, maintenance = manager.key(parsed_param, configuration)
        print(key_generated)

    except KNOWN_EXCEPTION as e:
        logger.error(f"{e}")
        logger.debug("", exc_info=True)
        sys.exit(-1)
    except Exception as e:
        logger.error(f"Error occurred while generating key from {parameters}, " f"{e}")
        logger.debug("", exc_info=True)
        sys.exit(-1)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("parameters", required=True)
@user_config_setup
@notification_server_setup
def value(parameters: str, configuration: conf.UserConfig):
    """
    Return the value on the server corresponding to the key which is generated according to the current schema and
    the parameters defined

    :param parameters: key1=value1,key2=value2,...
    """

    try:
        parsed_param = _parse_inline_params(parameters)
        v = manager.value(parsed_param, configuration)
        print(v)

    except KNOWN_EXCEPTION as e:
        logger.error(f"{e}")
        logger.debug("", exc_info=True)
        sys.exit(-1)
    except Exception as e:
        logger.error(f"Error occurred while return value for {parameters}, " f"{e}")
        logger.debug("", exc_info=True)
        sys.exit(-1)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("parameters", required=True)
@user_config_setup
@notification_server_setup
def notify(parameters: str, configuration: conf.UserConfig):
    """
    Create a notification with the parameters passed and submit it to the notification server
    :param parameters: key1=value1,key2=value2,...
    """

    try:
        parsed_param = _parse_inline_params(parameters)
        manager.notify(parsed_param, config=configuration)
        print("Done")

    except KNOWN_EXCEPTION as e:
        logger.error(f"{e}")
        logger.debug("", exc_info=True)
        sys.exit(-1)
    except Exception as e:
        logger.error(f"Error occurred while notifying the notification {parameters}, " f"{e}")
        logger.debug("", exc_info=True)
        sys.exit(-1)


cli.add_command(listen)
cli.add_command(key)
cli.add_command(value)
cli.add_command(notify)

if __name__ == "__main__":
    listen()


def _parse_inline_params(params: str) -> Dict[str, any]:
    """
    This helper method parses the notification string in a dictionary
    :param params:
    :return: notification as dictionary
    """
    logger.debug("Parsing the inline parameters...")
    parsed_param = {}
    ps = params.split(",")
    assert len(ps) > 1, "Wrong structure for the notification string, it should be <key_name>=<key_value>,..."
    for p in ps:
        pair = p.split("=")
        assert len(pair) == 2, "Wrong structure for the notification string, it should be <key_name>=<key_value>,..."
        parsed_param[pair[0]] = pair[1]
    logger.debug("Notification string successfully parsed")
    return parsed_param



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from .. import logger
from ..user_config import UserConfig
from .auth import Auth


class EtcdAuth(Auth):
    """
    This class set the authentication credentials for the basic etcd authentication where username and password are
    needed and once authenticated only the token is needed.
    """

    def __init__(self, config: UserConfig):
        super(EtcdAuth, self).__init__()
        logger.debug("Setting password from key file")
        self._password = config.password
        self._username = config.username
        logger.debug(f"Username: {self.username}")
        self._token = None  # this will be set once authenticated

    @property
    def username(self) -> str:
        return self._username

    @property
    def password(self) -> str:
        return self._password

    @property
    def token(self) -> str:
        return self._token

    @token.setter
    def token(self, token: str):
        self._token = token

    def header(self):
        header = {}
        if self.token:
            header["Authorization"] = self.token
        return header



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

__all__ = ["engine", "engine_factory", "etcd_grpc_engine", "etcd_rest_engine", "file_based_engine", "EngineType"]

import importlib
from enum import Enum


class EngineType(Enum):
    """
    This Enum describes the various notification mechanism currently available. This identifies the type of server
    this client is connecting to.
    """

    ETCD_GRPC = ("etcd_grpc_engine", "EtcdGrpcEngine")
    ETCD_REST = ("etcd_rest_engine", "EtcdRestEngine")
    FILE_BASED = ("file_based_engine", "FileBasedEngine")

    def __str__(self):
        return self.name.lower()

    def get_class(self):
        module = importlib.import_module("pyaviso.engine." + self.value[0])
        return getattr(module, self.value[1])



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import atexit
import collections.abc
import logging
import logging.config
import logging.handlers
import os
import re
import sys
from typing import Dict, Optional

import yaml

from . import HOME_FOLDER, SYSTEM_FOLDER, logger
from .authentication import AuthType
from .engine import EngineType
from .event_listeners.listener_schema_parser import ListenerSchemaParserType

# Default configuration location
CONF_FILE = "config.yaml"
KEY_FILE = "key"


class EngineConfig:
    def __init__(
        self,
        host: str,
        port: int,
        type: str,
        polling_interval: Optional[int] = None,
        max_file_size: Optional[int] = None,
        timeout: Optional[int] = None,
        service: Optional[str] = None,
        https: bool = False,
        catchup: Optional[bool] = None,
        automatic_retry_delay: Optional[int] = None,
    ):
        """
        :param host: endpoint host of the notification server
        :param port: endpoint port of the notification server
        :param type: interface to use to communicate to the notification and configuration servers
        :param polling_interval: interval between consecutive requests for new notifications, in seconds
        :param max_file_size: max file size allowed to push to the configuration server, in KiB
        :param timeout: number of seconds of waiting before timing-out the request to the server
        :param service: location in the configuration server associated to the notification service
        for the event listeners validation
        :param https: if True the connection will go through HTTPS
        :param catchup: if True the notification engine will first look for the missed notifications
        :param automatic_retry_delay: Number of seconds to wait before retrying to connect to the engine
        """
        self.host = host
        self.port = port
        self.type = EngineType[type.upper()]
        self.polling_interval = polling_interval
        self.max_file_size = max_file_size
        self.timeout = timeout
        self.https = https
        self.service = service
        self.catchup = catchup
        self.automatic_retry_delay = automatic_retry_delay

    def __str__(self):
        config_string = (
            f"host: {self.host}"
            + f", port: {self.port}"
            + f", https: {self.https}"
            + f", type: {self.type.name}"
            + f", polling_interval: {self.polling_interval}"
            + f", timeout: {self.timeout}"
            + f", max_file_size: {self.max_file_size}"
            + f", service: {self.service}"
            + f", catchup: {self.catchup}"
            + f", automatic_retry_delay: {self.automatic_retry_delay}"
        )
        return config_string


class UserConfig:
    """
    This class is in charge of holding the user configuration, which can be defined by command line options,
    environment variables, configuration files or defaults.
    """

    def __init__(
        self,
        conf_path: Optional[str] = None,
        conf_from_file: Optional[Dict[str, any]] = None,
        logging_path: Optional[str] = None,
        notification_engine: Optional[Dict[str, any]] = None,
        configuration_engine: Optional[Dict[str, any]] = None,
        debug: Optional[bool] = None,
        quiet: Optional[bool] = None,
        no_fail: Optional[bool] = None,
        username: Optional[str] = None,
        username_file: Optional[str] = None,
        key_file: Optional[str] = None,
        auth_type: Optional[str] = None,
        key_ttl: Optional[int] = None,
        schema_parser: Optional[str] = None,
        remote_schema: Optional[bool] = None,
        listeners: Optional[Dict[str, any]] = None,
    ):
        """
        :param conf_path: path to the system configuration file. If not provided,
        the default location is HOME_FOLDER/config.yaml.
        :param conf_from_file: system configuration dictionary. This will be treated as if read from file, in the same
        priority order.
        :param logging_path: path to the logging configuration file. If not provided,
        the default location is the logging section of the HOME_FOLDER/config.yaml.
        :param notification_engine: configuration object for the notification server
        :param configuration_engine: configuration object for the configuration server
        :param debug: flag to activate the debug log to the console output
        :param quiet: flag to activate only warning and error log to the console output
        :param no_fail: flag to suppress any error exit code
        :param username: username required to authenticate to the notification and configuration servers
        :param username_file: file path containing the username required to authenticate the user
        :param key_file: file path to the key required to authenticate to the notification and configuration servers
        :param auth_type: Authentication type
        :param key_ttl: Time to live of the keys submitted
        :param schema_parser: parser to use to build the listener schema
        :param remote_schema: flag to activate the dynamic retrieval of the listener schema from the configuration
        server
        :param listeners: listeners configuration
        """
        try:
            # we build the configuration in priority order from the lower to the higher
            # start from the defaults
            self._config = self._create_default_config()

            # add the configuration files
            if conf_from_file:  # only one method is allowed
                UserConfig.deep_update(self._config, conf_from_file)
            else:
                UserConfig.deep_update(self._config, self._parse_config_files(conf_path))

            # initialise logger, this needs to be done ASAP
            self.logging_setup(logging_path)

            # add environment variables
            UserConfig.deep_update(self._config, self._read_env_variables())

            # add constructor parameters
            self.notification_engine = notification_engine
            self.configuration_engine = configuration_engine
            self.debug = debug
            self.quiet = quiet
            self.no_fail = no_fail
            self.username = username
            self.username_file = username_file
            self.auth_type = auth_type
            if self.auth_type != AuthType.NONE:
                self.key_file = key_file
                self.password = self._read_key()
                if self.username_file:
                    self.username = self._read_username_file()
            self.key_ttl = key_ttl
            self.schema_parser = schema_parser
            self.remote_schema = remote_schema
            self.listeners = listeners

            logger.debug("Loading configuration completed")

        except Exception as e:
            logger.error(f"Error occurred while setting the configuration, {type(e)}, {e}")
            logger.debug("", exc_info=True)
            sys.exit(-1)

    @staticmethod
    def _create_default_config() -> Dict[str, any]:
        # notification engine
        notification_engine = {}
        notification_engine["host"] = "localhost"
        notification_engine["port"] = 2379
        notification_engine["https"] = False
        notification_engine["type"] = "etcd_rest"
        notification_engine["polling_interval"] = 30  # seconds
        notification_engine["timeout"] = 60  # seconds
        notification_engine["service"] = "aviso/v1"
        notification_engine["catchup"] = True
        notification_engine["automatic_retry_delay"] = 15  # seconds

        # configuration engine
        configuration_engine = {}
        configuration_engine["host"] = "localhost"
        configuration_engine["port"] = 2379
        configuration_engine["https"] = False
        configuration_engine["type"] = "etcd_rest"
        configuration_engine["max_file_size"] = 500  # KiB
        configuration_engine["timeout"] = 60  # seconds
        configuration_engine["automatic_retry_delay"] = 15  # seconds

        # main config
        config = {}
        config["notification_engine"] = notification_engine
        config["configuration_engine"] = configuration_engine
        config["username"] = None
        config["username_file"] = None
        config["debug"] = False
        config["quiet"] = False
        config["no_fail"] = False
        config["key_file"] = os.path.join(SYSTEM_FOLDER, KEY_FILE)
        config["auth_type"] = "none"
        config["key_ttl"] = -1  # not expiring
        config["schema_parser"] = "generic"
        config["remote_schema"] = False
        return config

    def _read_key(self) -> str:
        assert self.key_file is not None, "Key file not found"
        full_key_path = os.path.expanduser(self.key_file)
        try:
            with open(full_key_path, "r") as k:
                return k.read().rstrip()
        except FileNotFoundError as e:
            logger.error(f"Not able to load the key file: {full_key_path},  {e}")
            logger.debug("", exc_info=True)
            sys.exit(-1)

    def _read_username_file(self) -> str:
        assert self.username_file is not None, "Username file not found"
        full_username_path = os.path.expanduser(self.username_file)
        try:
            with open(full_username_path, "r") as u:
                return u.read().rstrip()
        except FileNotFoundError as e:
            logger.error(f"Not able to load the username file: {full_username_path},  {e}")
            logger.debug("", exc_info=True)
            sys.exit(-1)

    def _parse_config_files(self, user_conf_path: str) -> Dict[str, any]:
        # build the configuration dictionary from system and user inputs
        current_config = {}

        def parse_config(file_path: str):
            try:
                with open(file_path, "r") as f:
                    config = yaml.load(f.read(), Loader=HomeFolderLoader)
                # merge with the current config
                UserConfig.deep_update(current_config, config)
            except Exception as e:
                logger.error(f"Not able to load the configuration in {file_path},  {e}")
                logger.debug("", exc_info=True)
                sys.exit(-1)

        # First the system config file
        system_path = os.path.join(SYSTEM_FOLDER, CONF_FILE)
        # Check the directory exist
        if os.path.exists(system_path):
            parse_config(system_path)
        else:
            logger.debug(f"Configuration in {system_path} not found")

        # Second the Home config file
        home_path = os.path.join(os.path.expanduser(HOME_FOLDER), CONF_FILE)
        # Check the directory exist
        if os.path.exists(home_path):
            parse_config(home_path)
        else:
            logger.debug(f"Configuration in {home_path} not found")

        # Third the env variable
        if "AVISO_CONFIG" in os.environ:
            env_path = os.environ["AVISO_CONFIG"]
            parse_config(env_path)

        # Finally the user config option
        if user_conf_path:
            parse_config(user_conf_path)

        return current_config

    def _read_env_variables(self) -> Dict[str, any]:
        config = {"notification_engine": {}, "configuration_engine": {}}
        if "AVISO_NOTIFICATION_HOST" in os.environ:
            config["notification_engine"]["host"] = os.environ["AVISO_NOTIFICATION_HOST"]
        if "AVISO_NOTIFICATION_PORT" in os.environ:
            config["notification_engine"]["port"] = int(os.environ["AVISO_NOTIFICATION_PORT"])
        if "AVISO_NOTIFICATION_HTTPS" in os.environ:
            config["notification_engine"]["https"] = os.environ["AVISO_NOTIFICATION_HTTPS"]
        if "AVISO_NOTIFICATION_ENGINE" in os.environ:
            config["notification_engine"]["type"] = os.environ["AVISO_NOTIFICATION_ENGINE"]
        if "AVISO_NOTIFICATION_SERVICE" in os.environ:
            config["notification_engine"]["service"] = os.environ["AVISO_NOTIFICATION_SERVICE"]
        if "AVISO_NOTIFICATION_CATCHUP" in os.environ:
            config["notification_engine"]["catchup"] = os.environ["AVISO_NOTIFICATION_CATCHUP"]
        if "AVISO_POLLING_INTERVAL" in os.environ:
            config["notification_engine"]["polling_interval"] = int(os.environ["AVISO_POLLING_INTERVAL"])
        if "AVISO_CONFIGURATION_HOST" in os.environ:
            config["configuration_engine"]["host"] = os.environ["AVISO_CONFIGURATION_HOST"]
        if "AVISO_CONFIGURATION_PORT" in os.environ:
            config["configuration_engine"]["port"] = int(os.environ["AVISO_CONFIGURATION_PORT"])
        if "AVISO_CONFIGURATION_HTTPS" in os.environ:
            config["configuration_engine"]["https"] = os.environ["AVISO_CONFIGURATION_HTTPS"]
        if "AVISO_CONFIGURATION_ENGINE" in os.environ:
            config["configuration_engine"]["type"] = os.environ["AVISO_CONFIGURATION_ENGINE"]
        if "AVISO_MAX_FILE_SIZE" in os.environ:
            config["configuration_engine"]["max_file_size"] = int(os.environ["AVISO_MAX_FILE_SIZE"])
        if "AVISO_USERNAME" in os.environ:
            config["username"] = os.environ["AVISO_USERNAME"]
        if "AVISO_USERNAME_FILE" in os.environ:
            config["username_file"] = os.environ["AVISO_USERNAME_FILE"]
        if "AVISO_DEBUG" in os.environ:
            config["debug"] = os.environ["AVISO_DEBUG"]
        if "AVISO_QUIET" in os.environ:
            config["quiet"] = os.environ["AVISO_QUIET"]
        if "AVISO_NO_FAIL" in os.environ:
            config["no_fail"] = os.environ["AVISO_NO_FAIL"]
        if "AVISO_KEY_FILE" in os.environ:
            config["key_file"] = os.environ["AVISO_KEY_FILE"]
        if "AVISO_KEY_TTL" in os.environ:
            config["key_ttl"] = int(os.environ["AVISO_KEY_TTL"])
        if "AVISO_AUTH_TYPE" in os.environ:
            config["auth_type"] = os.environ["AVISO_AUTH_TYPE"]
        if "AVISO_REMOTE_SCHEMA" in os.environ:
            config["remote_schema"] = os.environ["AVISO_REMOTE_SCHEMA"]
        if "AVISO_SCHEMA_PARSER" in os.environ:
            config["schema_parser"] = os.environ["AVISO_SCHEMA_PARSER"]
        if "AVISO_TIMEOUT" in os.environ:  # one variable for both engine
            timeout = None if os.environ["AVISO_TIMEOUT"] == "null" else int(os.environ["AVISO_TIMEOUT"])
            config["notification_engine"]["timeout"] = timeout
            config["configuration_engine"]["timeout"] = timeout
        if "AVISO_AUTOMATIC_RETRY_DELAY" in os.environ:  # one variable for both engine
            automatic_retry_delay = (
                None
                if os.environ["AVISO_AUTOMATIC_RETRY_DELAY"] == "null"
                else int(os.environ["AVISO_AUTOMATIC_RETRY_DELAY"])
            )
            config["notification_engine"]["automatic_retry_delay"] = automatic_retry_delay
            config["configuration_engine"]["automatic_retry_delay"] = automatic_retry_delay
        return config

    def logging_setup(self, logging_conf_path: str):

        if logging_conf_path is not None:
            try:
                with open(logging_conf_path, "r") as f:
                    log_config = yaml.safe_load(f.read())
            except Exception as e:
                logger.warning(f"Not able to load the logging configuration,  {e}")
                logger.debug("", exc_info=True)
                sys.exit(-1)
        elif "AVISO_LOG" in os.environ:
            try:
                with open(os.environ["AVISO_LOG"], "r") as f:
                    log_config = yaml.safe_load(f.read())
            except Exception as e:
                logger.warning(f"Not able to load the logging configuration,  {e}")
                logger.debug("", exc_info=True)
                sys.exit(-1)
        elif self._config is not None and self._config.get("logging") is not None:
            # Setting logging from user configuration file
            log_config = self._config.get("logging")
        else:  # Defaults
            # Configure the logging with the default configuration
            self._configure_default_log()
            return

        # initialise the logging with the user configuration
        try:
            logging.config.dictConfig(log_config)
        except Exception as e:
            logger.warning(f"Not able to load the logging configuration, {e}")
            logger.debug("", exc_info=True)
            sys.exit(-1)

    @property
    def notification_engine(self) -> EngineConfig:
        return self._notification_engine

    @notification_engine.setter
    def notification_engine(self, notification_engine: Dict[str, any]):
        ne = self._config.get("notification_engine")
        if notification_engine is not None and ne is not None:
            UserConfig.deep_update(ne, notification_engine)
        elif notification_engine is not None:
            ne = notification_engine
        # verify is valid
        assert ne is not None, "notification_engine has not been configured"
        assert "host" in ne, "notification_engine host has not been configured"
        assert "port" in ne, "notification_engine port has not been configured"
        assert "type" in ne, "notification_engine type has not been configured"
        assert "https" in ne, "notification_engine https has not been configured"
        assert "service" in ne, "notification_engine service has not been configured"
        assert "catchup" in ne, "notification_engine catchup has not been configured"
        assert "automatic_retry_delay" in ne, "notification_engine automatic_retry_delay has not been configured"
        assert "polling_interval" in ne, "notification_engine polling_interval has not been configured"
        assert "timeout" in ne, "notification_engine timeout has not been configured"
        if type(ne["https"]) is str:
            ne["https"] = ne["https"].casefold() == "true".casefold()
        if type(ne["catchup"]) is str:
            ne["catchup"] = ne["catchup"].casefold() == "true".casefold()

        # translate the ne in a NotificationEngineConfig
        self._notification_engine = EngineConfig(
            ne["host"],
            ne["port"],
            ne["type"],
            polling_interval=ne["polling_interval"],
            timeout=ne["timeout"],
            https=ne["https"],
            service=ne["service"],
            catchup=ne["catchup"],
            automatic_retry_delay=ne["automatic_retry_delay"],
        )

    @property
    def configuration_engine(self) -> EngineConfig:
        return self._configuration_engine

    @configuration_engine.setter
    def configuration_engine(self, configuration_engine: Dict[str, any]):
        ce = self._config.get("configuration_engine")
        if configuration_engine is not None and ce is not None:
            UserConfig.deep_update(ce, configuration_engine)
        elif configuration_engine is not None:
            ce = configuration_engine
        # verify is valid
        assert ce is not None, "configuration_engine has not been configured"
        assert "host" in ce, "configuration_engine host has not been configured"
        assert "port" in ce, "configuration_engine port has not been configured"
        assert "type" in ce, "configuration_engine type has not been configured"
        assert "https" in ce, "configuration_engine https has not been configured"
        assert "max_file_size" in ce, "configuration_engine max_file_size has not been configured"
        assert "timeout" in ce, "configuration_engine timeout has not been configured"
        assert "automatic_retry_delay" in ce, "configuration_engine automatic_retry_delay has not been configured"
        if type(ce["https"]) is str:
            ce["https"] = ce["https"].casefold() == "true".casefold()

        # exclude file_based from the options for the configuration engine
        assert ce["type"].casefold() != "file_based", "File_based engine not available as configuration engine"
        # translate the ce in a ConfigurationEngineConfig
        self._configuration_engine = EngineConfig(
            ce["host"],
            ce["port"],
            ce["type"],
            max_file_size=ce["max_file_size"],
            timeout=ce["timeout"],
            https=ce["https"],
            automatic_retry_delay=ce["automatic_retry_delay"],
        )

    @property
    def schema_parser(self) -> ListenerSchemaParserType:
        return self._schema_parser

    @schema_parser.setter
    def schema_parser(self, schema_parser: str):
        self._schema_parser = ListenerSchemaParserType[self._configure_property(schema_parser, "schema_parser").upper()]

    @property
    def remote_schema(self) -> bool:
        return self._remote_schema

    @remote_schema.setter
    def remote_schema(self, remote_schema: str):
        self._remote_schema = self._configure_property(remote_schema, "remote_schema")
        if type(self._remote_schema) is str:
            self._remote_schema = self._remote_schema.casefold() == "true".casefold()

    @property
    def key_ttl(self):
        return self._key_ttl

    @key_ttl.setter
    def key_ttl(self, key_ttl: int):
        self._key_ttl = self._configure_property(key_ttl, "key_ttl")

    @property
    def username(self):
        return self._username

    @username.setter
    def username(self, username: str):
        self._username = self._configure_property(username, "username", nullable=True)

    @property
    def username_file(self) -> str:
        return self._username_file

    @username_file.setter
    def username_file(self, username_file: str):
        self._username_file = self._configure_property(username_file, "username_file", nullable=True)

    @property
    def key_file(self) -> str:
        return self._key_file

    @key_file.setter
    def key_file(self, key_file: str):
        self._key_file = self._configure_property(key_file, "key_file")

    @property
    def auth_type(self) -> AuthType:
        return self._auth_type

    @auth_type.setter
    def auth_type(self, auth_type: str):
        self._auth_type = AuthType[self._configure_property(auth_type, "auth_type").upper()]

    @property
    def listeners(self) -> Dict:
        return self._listeners

    @listeners.setter
    def listeners(self, listeners: Dict):
        self._listeners = {"listeners": self._configure_property(listeners, "listeners", nullable=True)}

    @property
    def no_fail(self) -> bool:
        return self._no_fail

    @no_fail.setter
    def no_fail(self, no_fail: str):
        self._no_fail = self._configure_property(no_fail, "no_fail")
        if type(self._no_fail) is str:
            self._no_fail = self._no_fail.casefold() == "true".casefold()
        if self.no_fail:
            # define a function to run at exit
            def suppress_exit_code():
                # override any other exit code and exit
                os._exit(0)

            atexit.register(suppress_exit_code)

    @property
    def debug(self) -> bool:
        return self._debug

    @debug.setter
    def debug(self, debug: any):
        self._debug = self._configure_property(debug, "debug")
        if type(self._debug) is str:
            self._debug = self._debug.casefold() == "true".casefold()
        if self._debug:
            logging_level = logging.DEBUG
            # set the root level
            logging.root.setLevel(logging_level)
            # Configuring console logging
            try:
                console = next(h for h in logging.getLogger().handlers if h.name == "console")
                console.setLevel(logging_level)
            except StopIteration:  # this is raised when the console logger could not be found
                # set the general logger - Note this will affect also the logging file
                logging.getLogger().setLevel(logging_level)

    @property
    def quiet(self) -> bool:
        return self._quiet

    @quiet.setter
    def quiet(self, quiet: any):
        self._quiet = self._configure_property(quiet, "quiet")
        if type(self._quiet) is str:
            self._quiet = self._quiet.casefold() == "true".casefold()
        if self._quiet:
            logging_level = logging.ERROR
            # Configuring console logging
            try:
                console = next(h for h in logging.getLogger().handlers if h.name == "console")
                console.setLevel(logging_level)
            except StopIteration:  # this is raised when the console logger could not be found
                # ignore it in this case
                pass

    def __str__(self):
        config_string = (
            f"notification_engine: {self.notification_engine}"
            + f", configuration_engine: {self.configuration_engine}"
            + f", auth_type: {self.auth_type}"
            + f", debug: {self.debug}"
            + f", quiet: {self.quiet}"
            + f", username: {self.username}"
            + f", username_file: {self.username_file}"
            + f", no_fail: {self.no_fail}"
            + f", key_ttl: {self.key_ttl}"
            + f", schema_parser: {self.schema_parser}"
            + f", remote_schema: {self.remote_schema}"
        )
        return config_string

    def _configure_default_log(self):
        try:
            next(h for h in logging.getLogger().handlers if h.name == "console")
            # don't do anything, we already have a console handler
        except StopIteration:  # this is raised when the console logger could not be found
            # creating default console handler
            console_handler = logging.StreamHandler()
            console_handler.name = "console"
            console_handler.setLevel(logging.INFO)
            console_handler.setFormatter(logging.Formatter("%(message)s"))
            logging.getLogger().addHandler(console_handler)

    def _configure_property(self, param, name, nullable=False):
        value = None
        if param is not None:
            value = param
        elif self._config.get(name) is not None:
            # Setting var from user configuration file
            value = self._config.get(name)
        else:
            if not nullable:
                logger.error(f"{name} has not been configured")
                sys.exit(-1)
        return value

    @staticmethod
    def deep_update(d, u):
        for k, v in u.items():
            if isinstance(v, collections.abc.Mapping):
                d[k] = UserConfig.deep_update(d.get(k, type(v)()), v)
            else:
                d[k] = v
        return d


# class to allow yaml loader to replace ~ with HOME directory
class HomeFolderLoader(yaml.SafeLoader):
    path_matcher = re.compile("~")

    @staticmethod
    def path_constructor(loader, node):
        return os.path.expanduser(node.value)


HomeFolderLoader.add_implicit_resolver("!path", HomeFolderLoader.path_matcher, None)
HomeFolderLoader.add_constructor("!path", HomeFolderLoader.path_constructor)



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from typing import List

from .type_handler import TypeHandler


class IntHandler(TypeHandler):
    def __init__(self, key, required=False, canonic=None, range=None):
        super(IntHandler, self).__init__(key, required)
        self._canonic = canonic
        self._range = range

    @property
    def canonic(self) -> str:
        return self._canonic

    @property
    def range(self) -> List[int]:
        return self._range

    def valid(self, value: any) -> bool:
        try:
            value = int(value)
        except ValueError as e:
            raise ValueError(f"Key {self.key} has to be an integer", e)
        if self.range is not None:
            assert len(self.range) == 2, "Wrong schema structure, range can only have 2 elements"
            if value < self.range[0] or value > self.range[1]:
                raise ValueError(f"Value {value} for key {self.key} is outside the range defined")
        return True

    def canonise(self, value: any) -> str:
        value = int(value)
        if self.canonic is not None:
            return self.canonic.format(value)
        else:
            return str(value)



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import fcntl
import json
import os
import time
from abc import ABC, abstractmethod
from datetime import datetime
from queue import Queue
from typing import Any, Tuple

from .. import HOME_FOLDER, logger
from ..authentication.auth import Auth
from ..custom_exceptions import EngineException, EngineHistoryNotAvailableError
from ..user_config import EngineConfig
from .engine import DATE_FORMAT, Engine

MAX_KV_RETURNED = 10000
LOCAL_STATE_FOLDER = "etcd/last"
LAST_REVISION_FILE = "revision.json"


class EtcdEngine(Engine, ABC):
    """
    This class is a specialisation of the Engine class but it is still abstract. It provides common functions for the
    Etcd specialisation.
    """

    def __init__(self, config: EngineConfig, auth: Auth):
        super(EtcdEngine, self).__init__(config, auth)

    @abstractmethod
    def _latest_revision(self, key: str) -> int:
        """
        :param: key used for the server request
        :return: latest revision of the notification server.
        """
        pass

    @abstractmethod
    def _lease(self, ttl) -> str:
        """
        This method requests a Lease for the TTL specified
        :param ttl: Lease TTL
        :return: lease id
        """
        pass

    def _polling(
        self,
        key: str,
        callback: callable([str, str]),
        channel: Queue,
        from_date: datetime = None,
        to_date: datetime = None,
    ):
        """
        This method implements the active polling
        :param key: key to watch as a prefix
        :param callback: function to call if any change happen
        :param channel: global communication channel among threads
        :param from_date: date from when to request notifications, if None it will be from now
        :param to_date: date until when to request notifications, if None it will be until now
        :return:
        """

        def trigger_callback(notifications):
            for notification in notifications:
                v = notification["value"].decode()
                k = notification["key"]
                logger.debug(f"Notification received for key {k}")
                try:
                    callback(k, v)
                except Exception as err:
                    logger.error(f"Error with notification trigger: {err}")
                    logger.debug("", exc_info=True)

        try:
            # initialise the revisions
            final_rev = None

            # check start date
            if from_date is None:  # no start date defined
                if self.catchup is None:
                    raise EngineException("catchup not defined for notification engine")
                if self.catchup:  # we start from the saved one
                    saved_rev = self._last_saved_revision()
                    if saved_rev != -1:
                        logger.info("Starting from last notification received")
                        next_rev = saved_rev
                    else:  # if it's the first time we start from now
                        next_rev = self._latest_revision(key) + 1
                else:  # delete the saved state
                    self._delete_saved_revision()
                    # we start from now
                    next_rev = self._latest_revision(key) + 1

            else:  # start date defined
                logger.info("Searching for past notifications...")
                next_rev, final_rev = self._from_to_revisions(key, from_date=from_date, to_date=to_date)
                if next_rev == -1 and final_rev == -1:
                    logger.warning("No history available in the time period selected")
                    channel.put(True)
                    return
                elif next_rev:
                    logger.info("Search completed, retrieving...")
                else:
                    logger.error("Error in one of the listening process")
                    channel.put(False)
                    return

            # check end date
            if to_date:  # end date defined, retrieve only past notifications
                if final_rev:
                    kvs = self.pull(key, min_rev=next_rev, max_rev=final_rev)
                    # remove the status from the result
                    for kv in kvs:
                        if kv["key"] == key:  # this is the status
                            kvs.remove(kv)
                            break
                    # trigger the callback
                    trigger_callback(kvs)
                # de-register this pooling thread as we have finished
                self.stop(key)
                logger.info("Search and retrieval completed")
                if len(self._listeners) == 0:  # this is the last pooling thread
                    # terminate the main execution
                    channel.put(True)
                    return

            else:  # no end date defined, start the polling for new notifications
                while key in self._listeners:  # this is the stop condition
                    # retrieve any change since the last revision
                    kvs = self.pull(key, min_rev=next_rev)
                    # remove the status from the result
                    for kv in kvs:
                        if kv["key"] == key:  # this is the status
                            kvs.remove(kv)
                            break
                    if len(kvs) > 0:
                        # update the current revision
                        for kv in kvs:
                            if next_rev < kv["mod_rev"] + 1:
                                next_rev = kv["mod_rev"] + 1
                        # save current rev
                        self._save_last_revision(next_rev)
                        # trigger the callback
                        trigger_callback(kvs)
                    # wait the polling interval before trying again
                    time.sleep(self._polling_interval)

        except Exception as e:
            logger.error(f"Error while listening to key {key}: {e}")
            logger.debug("", exc_info=True)
            channel.put(False)

    def _last_saved_revision(self) -> int:
        """
        This method is used to read the last revision saved to file in the home folder
        :return: last revision or -1 if no revision could be read
        """
        # build the path where the last revision is saved
        full_home_path = os.path.expanduser(HOME_FOLDER)
        full_rev_path = os.path.join(full_home_path, LOCAL_STATE_FOLDER, LAST_REVISION_FILE)
        if os.path.exists(full_rev_path):
            try:
                with open(full_rev_path, "r") as f:
                    # acquire a file lock to avoid concurrency among processes
                    fcntl.lockf(f, fcntl.LOCK_SH)
                    rev_dict = json.loads(f.read())
                    last_rev = rev_dict["last_revision"]
                    # release file lock
                    fcntl.lockf(f, fcntl.LOCK_UN)
                logger.debug(f"Last revision saved is {last_rev}")
                return last_rev
            except Exception as e:
                logger.warning(f"Error occurred while reading the last revision saved: {e}")
                logger.debug("", exc_info=True)

        # default return
        return -1

    def _delete_saved_revision(self):
        """
        This method is used to delete the file where the last revision is saved
        """
        # build the path where the last revision is saved
        full_home_path = os.path.expanduser(HOME_FOLDER)
        full_state_path = os.path.join(full_home_path, LOCAL_STATE_FOLDER)
        full_rev_path = os.path.join(full_state_path, LAST_REVISION_FILE)

        with self._state_lock:  # multiple listing threads could access this simultaneously
            if os.path.exists(full_rev_path):
                try:
                    # delete the file
                    os.remove(full_rev_path)
                    logger.debug("Last revision file has been successfully deleted")
                except Exception:
                    logger.warning(f"Deleting the last revision file has failed: {full_rev_path}")
                    logger.debug("", exc_info=True)
                    return False

    def _save_last_revision(self, rev: int) -> bool:
        """
        This method is used to save to file the revision passed as the last revision pulled
        :param rev: last revision to save
        :return: True if saved otherwise False
        """
        if rev is not None:
            rev_dict = {
                "last_revision": rev,
                "date_time": datetime.utcnow().strftime(DATE_FORMAT),
                "server_host": self.host,
                "server_port": self.port,
            }
            # build the path where the last revision will be saved
            full_home_path = os.path.expanduser(HOME_FOLDER)
            full_state_path = os.path.join(full_home_path, LOCAL_STATE_FOLDER)

            with self._state_lock:  # multiple listing threads could access this simultaneously
                if not os.path.exists(full_state_path):
                    try:
                        os.makedirs(full_state_path, exist_ok=True)
                    except OSError:
                        logger.warning(f"Cannot create the directory for the last revision: {full_state_path}")
                        logger.debug("", exc_info=True)
                        return False
                try:
                    full_rev_path = os.path.join(full_state_path, LAST_REVISION_FILE)
                    with open(full_rev_path, "w") as f:
                        # acquire a file lock to avoid concurrency among processes
                        fcntl.lockf(f, fcntl.LOCK_EX)
                        # write the revision in the file
                        json.dump(rev_dict, f)
                        # release file lock
                        fcntl.lockf(f, fcntl.LOCK_UN)
                    logger.debug(f"Last revision pulled {rev} has been successfully saved")
                except Exception:
                    logger.warning(f"Saving of the last revision has failed: {full_state_path}")
                    logger.debug("", exc_info=True)
                    return False
        return True

    def _retrieve_status_history(self, key, rev=None) -> Tuple[int, Any, int, Any]:
        """
        :param key: key for which to return the status
        :param rev: revision of the status requested
        :return: a tuple: status's revision, status's date, previous status's revision,
        revision of the last status of the previous day
        """
        try:
            kvs = self.pull(key=key, prefix=False, rev=rev)
        # in case of retrieving a compacted revision we will get a 400 error with a proper reason
        except EngineHistoryNotAvailableError:
            logger.debug(f"Revision {rev} too old for current history")
            return rev, None, rev, None  # return in a way that is clear that we arrived at the end of the history

        # look inside what returned
        if len(kvs) == 0:
            return -1, None, -1, None
        assert len(kvs) == 1, f"Error in finding revision for {key}, more then one kv found or none"
        s_rev = kvs[0]["mod_rev"]
        s = json.loads(kvs[0]["value"].decode())
        if "prev_rev" in s:
            s_prev_rev = s["prev_rev"]
        else:
            s_prev_rev = -1  # this means that this is the first revision of the status. We cannot go further back
        s_date = datetime.strptime(s["date_time"], DATE_FORMAT)
        s_last_prev_day_rev = s.get("last_prev_day_rev")
        return s_rev, s_date, s_prev_rev, s_last_prev_day_rev

    def _from_to_revisions(self, key: str, from_date: datetime, to_date: datetime = None) -> Tuple[Any, Any]:
        """
        This methods search for revisions corresponding to the interval (from_date, to_date)
        :param key:
        :param from_date:
        :param to_date:
        :return: a tuple: revision just after from_date, revision just before to_date
        """
        logger.debug(f"Querying notification server to find historic revisions for key {key}")

        # first retrieve the current status
        to_rev = None
        status_rev, status_date, status_prev_rev, status_last_prev_day_rev = self._retrieve_status_history(key)

        # check all limit cases first
        # limit case - check if there is no status
        if status_rev == -1:
            logger.debug("No status available")
            return None, None
        # limit case - check if this revision is before from_date and to_date
        if status_date <= from_date:
            # return the revisions as this is the last point
            from_rev = status_rev + 1
            to_rev = from_rev
            return from_rev, to_rev
        else:
            # limit case - check if this revision is after from_date and to_date and it's the only point
            if status_prev_rev == -1 and to_date and status_date >= to_date:
                # return the revisions as this is the only point
                from_rev = status_rev - 1
                to_rev = from_rev
                return from_rev, to_rev
            # limit case - check if this revision is inside from_date and to_date and it's the only point
            elif status_prev_rev == -1:
                # return the revisions as this is the only point
                from_rev = status_rev
                to_rev = from_rev
                return from_rev, to_rev

        # if we have not returned yet search for older status revisions

        # check if we are inside the interval
        if to_date and status_date < to_date:
            to_rev = status_rev

        # start navigating the history
        while status_date > from_date and status_prev_rev != -1:
            if status_date.date() == from_date.date():  # same day
                # go back one revision
                status_rev, status_date, status_prev_rev, status_last_prev_day_rev = self._retrieve_status_history(
                    key, status_prev_rev
                )
            elif status_last_prev_day_rev:
                # go back to the revision of the last of the previous day -> we skip a day
                status_rev, status_date, status_prev_rev, status_last_prev_day_rev = self._retrieve_status_history(
                    key, status_last_prev_day_rev
                )
            elif status_last_prev_day_rev is None:  # it is the last point of history
                logger.warning("Reached the end of history available")
                break

            if status_date is None:  # if we reached a compacted revision status_date will be None,
                logger.warning("Reached the end of history available")
                break
            # check if this revision satisfies to_date and has not been previously saved
            if to_date and status_date < to_date and to_rev is None:
                to_rev = status_rev

        if status_date is None or status_date <= from_date:  # we exit the loop because we went out of the interval
            # save the revision in from_rev but increment it so we stay just inside the interval
            from_rev = status_rev + 1
        else:  # we exit because is the last point but we are inside the interval
            from_rev = status_rev
        if to_date and to_rev is None:  # this means there are no point inside the interval - limit case
            logger.debug("No keys found")
            return -1, -1

        return from_rev, to_rev

    def _incr_last_byte(self, path: str) -> bytes:
        """
        This function determines the end of the range required for a range call with the etcd3 API
        By incrementing the last byte of the input path, it allows to make a range call describing the input
        path as a branch rather than a leaf path.

        :param path: the path representing the start of the range
        :return: the path representing the end of the range
        """
        bytes_types = (bytes, bytearray)
        if not isinstance(path, bytes_types):
            if not isinstance(path, str):
                path = str(path)
            path = path.encode("utf-8")
        s = bytearray(path)
        # increment the last byte
        s[-1] = s[-1] + 1
        return bytes(s)



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import functools
import json
import sys

import click

from . import __version__, logger
from .cli_aviso import KNOWN_EXCEPTION, user_config_setup
from .service_config_manager import ServiceConfigManager
from .user_config import UserConfig

# main ServiceConfigManager object, we can use this for all commands

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


def configuration_server_setup(f):
    @click.option("--host", "-H", help="Configuration server host.")
    @click.option("--port", "-P", help="Configuration server port.", type=int)
    @functools.wraps(f)
    def functor(*args, **kwargs):
        if kwargs["host"] is not None:
            kwargs["configuration"].configuration_engine.host = kwargs["host"]
        kwargs.pop("host")
        if kwargs["port"] is not None:
            kwargs["configuration"].configuration_engine.port = kwargs["port"]
        kwargs.pop("port")

        return f(*args, **kwargs)

    return functor


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(version=__version__)
def cli():
    pass


@click.command(
    help="Push all files from the directory selected to the service defined, respecting the " "subdirectory structure."
)
@user_config_setup
@configuration_server_setup
@click.argument("service", required=True)
@click.option("--dir", "-D", help="Directory to push.", required=True)
@click.option("--message", "-m", help="Message to associate to the push.", required=True)
@click.option(
    "--delete", help="Allows delete of files on server if they don`t exist locally.", is_flag=True, default=False
)
def push(service: str, dir: str, message: str, delete: bool, configuration: UserConfig):
    config_manager = ServiceConfigManager(configuration)
    # call the push method of the ServiceConfigManager
    try:
        pushed_files = config_manager.push(service, dir, message, delete)
        for f in pushed_files:
            print(f)
        if len(pushed_files) > 0:
            logger.info(f"Push operation for service {service} successfully executed")
        else:
            logger.info(f"No file found to push for service {service}")

    except KNOWN_EXCEPTION as e:
        logger.error(f"{e}")
        logger.debug("", exc_info=True)
        sys.exit(-1)
    except Exception as e:
        logger.error(f"Error occurred while pushing the directory {dir} to the service {service}, " f"{e}")
        logger.debug("", exc_info=True)
        sys.exit(-1)


@click.command(help="Pull all files associated with the service defined.")
@user_config_setup
@configuration_server_setup
@click.argument("service", required=True)
@click.option("--dir", "-D", help="Directory to push.", default=".")
@click.option(
    "--delete", help="Allows delete of local files if they don`t exist on server.", is_flag=True, default=False
)
def pull(service: str, dir: str, delete: bool, configuration: UserConfig):
    config_manager = ServiceConfigManager(configuration)
    # call the pull_and_save method of the ServiceConfigManager
    try:
        pulled_files = config_manager.pull_and_save(service, dir, delete)
        for f in pulled_files:
            print(f)
        if len(pulled_files) > 0:
            logger.info(f"Pull operation for service {service} successfully executed")
        else:
            logger.info(f"No files found for service {service}")

    except KNOWN_EXCEPTION as e:
        logger.error(f"{e}")
        logger.debug("", exc_info=True)
        sys.exit(-1)
    except Exception as e:
        logger.error(f"Error occurred while pull the service {service}, " f"{e}")
        logger.debug("", exc_info=True)
        sys.exit(-1)


@click.command(help="Remove all files associated with the service defined.")
@user_config_setup
@configuration_server_setup
@click.argument("service", required=True)
@click.option("--doit", "-f", help="Remove without prompt.", is_flag=True, default=None)
def remove(service: str, doit: bool, configuration: UserConfig):
    config_manager = ServiceConfigManager(configuration)
    try:
        if not doit:
            # call the pull method of the ServiceConfigManager
            kvs = config_manager.pull(service, key_only=True)
            for kv in kvs:
                file_name = kv["key"]
                print(file_name)
            if len(kvs) > 0:
                logger.info("remove --doit to delete these files")
            else:
                logger.info(f"No files found for service {service}")
        else:  # delete all
            files_deleted = config_manager.remove(service)
            for f in files_deleted:
                print(f)
            if len(files_deleted) > 0:
                logger.info(f"Remove operation for service {service} successfully executed")
            else:
                logger.info(f"No files found for service {service}")

    except KNOWN_EXCEPTION as e:
        logger.error(f"{e}")
        logger.debug("", exc_info=True)
        sys.exit(-1)
    except Exception as e:
        logger.error(f"Error occurred while removing all files associated to the service {service}, " f"{e}")
        logger.debug("", exc_info=True)
        sys.exit(-1)


@click.command(help="Revert all files associated with the service defined to the previous version.")
@user_config_setup
@configuration_server_setup
@click.argument("service", required=True)
def revert(service: str, configuration: UserConfig):
    config_manager = ServiceConfigManager(configuration)
    # call the revert method of the ServiceConfigManager
    try:
        reverted_files = config_manager.revert(service)
        for f in reverted_files:
            print(f)
        if len(reverted_files) > 0:
            logger.info(f"Revert operation for service {service} successfully executed")
        else:
            logger.info(f"No files found for service {service}")

    except KNOWN_EXCEPTION as e:
        logger.error(f"{e}")
        logger.debug("", exc_info=True)
        sys.exit(-1)
    except Exception as e:
        logger.error(f"Error occurred while reverting the service {service}, " f"{e}")
        logger.debug("", exc_info=True)
        sys.exit(-1)


@click.command(help="Retrieve the status of the service defined.")
@user_config_setup
@configuration_server_setup
@click.argument("service", required=True)
def status(service: str, configuration: UserConfig):
    config_manager = ServiceConfigManager(configuration)
    # call the status method of the ServiceConfigManager
    try:
        s = config_manager.status(service)
        if s != {}:
            print(json.dumps(s, indent=4, sort_keys=True))

    except KNOWN_EXCEPTION as e:
        logger.error(f"{e}")
        logger.debug("", exc_info=True)
        sys.exit(-1)
    except Exception as e:
        logger.error(f"Error occurred while retrieving version for the service {service}, " f"{e}")
        logger.debug("", exc_info=True)
        sys.exit(-1)


cli.add_command(push)
cli.add_command(pull)
cli.add_command(remove)
cli.add_command(revert)
cli.add_command(status)

if __name__ == "__main__":
    cli()



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import datetime

from .type_handler import TypeHandler


class DateHandler(TypeHandler):
    def __init__(self, key, canonic, required=False):
        super(DateHandler, self).__init__(key, required)
        self._canonic = canonic

    @property
    def canonic(self) -> str:
        return self._canonic

    def valid(self, value: any) -> bool:
        try:
            self._d_datetime = datetime.datetime.strptime(str(value), self.canonic)
            return True
        except ValueError as e:
            raise ValueError("Date attribute is not complying with the format defined", e)

    def canonise(self, value: any) -> str:
        # strptime tolerates months or days with no leading zero, we need to format it again to be sure they are there
        return self._d_datetime.strftime(self.canonic)



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from typing import List

from .type_handler import TypeHandler


class EnumHandler(TypeHandler):
    def __init__(self, key, values: List[str], required=False, default=None):
        super(EnumHandler, self).__init__(key, required)
        self._valid_values = values
        self._default = default

    @property
    def valid_values(self) -> List[str]:
        return self._valid_values

    def valid(self, value: any) -> bool:
        if value == "" and self._default is not None:
            value = self._default
        # convert the value to the same type of the enums otherwise it will not be able to validate
        try:
            value = type(self.valid_values[0])(value)
        except ValueError as e:
            raise ValueError(f"Key {self.key} is not of a valid type", e)

        if value in self.valid_values:
            return True
        else:
            valid_values_str = ",".join(map(lambda x: str(x), self.valid_values))
            raise ValueError(f"Key {self.key} accepts only the following values: {valid_values_str}")

    def canonise(self, value: any) -> str:
        if value == "" and self._default is not None:
            value = self._default
        return value



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from typing import Callable, Dict

from .. import logger
from . import trigger
from .trigger import TriggerType


class FunctionTrigger(trigger.Trigger):
    """
    This class implements the 'Function' trigger by executing the function defined by the user and
    passing the notification key and value and the params as argument
    """

    def __init__(self, notification: Dict[str, any], params: Dict[str, any]):
        trigger.Trigger.__init__(self, notification, params)
        assert self.params.get("function") is not None, "'function' is a mandatory field for the 'Function' trigger"
        self.function: Callable = self.params.get("function")
        self.trigger_type = TriggerType.function

    def execute(self):
        logger.info("Starting Function Trigger...")
        logger.debug(f"calling function {self.function.__name__}")

        # run the function
        logger.debug("Running function trigger")
        self.function(self.notification)

        logger.info("Function Trigger completed")



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import re

from .type_handler import TypeHandler


class RegexHandler(TypeHandler):
    def __init__(self, key, regex, required=False):
        super(RegexHandler, self).__init__(key, required)
        self._regex = regex

    @property
    def regex(self) -> str:
        return self._regex

    def valid(self, value: any) -> bool:
        result = re.match(self.regex, value)
        if result is None:
            return False
        else:
            return True

    def canonise(self, value: any) -> str:
        return value



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from .date_handler import DateHandler
from .enum_handler import EnumHandler
from .float_handler import FloatHandler
from .int_handler import IntHandler
from .regex_handler import RegexHandler
from .string_handler import StringHandler
from .time_handler import TimeHandler
from .type_handler import TypeHandler

__all__ = [
    "DateHandler",
    "EnumHandler",
    "TimeHandler",
    "StringHandler",
    "IntHandler",
    "TypeHandler",
    "FloatHandler",
    "RegexHandler",
]



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from datetime import datetime
from typing import Dict, List

from .. import logger, user_config
from ..authentication.auth import Auth
from ..custom_exceptions import EventListenerException
from ..engine import engine_factory as ef
from . import event_listener_factory as elf
from .event_listener import EventListener


class ListenerManager:
    """
    This class manages the execution of the various event listeners
    """

    def __init__(self):
        self._listeners: List[EventListener] = []

    @property
    def listeners(self) -> List[EventListener]:
        return self._listeners

    def _run_listeners(self) -> bool:
        """
        This method is used to execute all the listeners currently managed

        :return: True if all the listeners are in execution, False otherwise
        """
        logger.debug("Calling run all listeners...")
        result = True
        listener_to_remove: List[EventListener] = []
        for listener in self._listeners:
            # Execute the listener
            if not listener.listen():
                result = False
                listener_to_remove.append(listener)
            else:
                keys = ",".join(listener.keys)
                logger.info(f"Listening to {keys} at {listener.engine.host}:{listener.engine.port}...")

        # now remove all of the listeners that were not able to start
        for listener in listener_to_remove:
            self._listeners.remove(listener)

        return result

    def _add_listener(self, listener: EventListener) -> None:
        """
        Add a listener to the internal listener list of the manager

        :param listener: EventListener object
        """
        self._listeners.append(listener)

    def _add_listeners(self, listeners: List[EventListener]) -> None:
        """
        Add the listener list to the internal listener list of the manager

        :param listeners: EventListener list
        """
        for listener in listeners:
            self._add_listener(listener)

    def _stop_listener(self, listener: EventListener) -> bool:
        """
        Stop the execution of listener passed as argument.

        :param listener: EventListener object
        :return: True if stopped, False otherwise
        """
        try:
            logger.debug(f"Calling stop {listener}...")
            # Stop the listener
            return listener.stop()
        except ValueError as error:
            logger.error(f"Error in stopping listener, exception message: {error}")
            logger.debug("", exc_info=True)
            return False

    def _cancel_listener(self, listener: EventListener) -> None:
        """
        Stop and delete the listener passed as argument

        :param listener: EventListener object
        """
        # first stop listener
        self._stop_listener(listener)
        # now remove it from the list
        self._listeners.remove(listener)

    def cancel_listeners(self) -> None:
        """
        Stop the execution of any listener currently in execution

        :return: True if no listener is currently in execution
        """
        # first cancel all the notification listeners
        for listener in self._listeners:
            self._stop_listener(listener)

        # now remove all of them from the internal list
        self._listeners.clear()

    def listen(
        self,
        listeners: List[Dict[str, any]],
        listener_schema: Dict[str, any],
        config: user_config.UserConfig = None,
        from_date: datetime = None,
        to_date: datetime = None,
    ) -> int:
        """
        This method implements the main workflow to instantiate and execute new listeners
        :param listeners: listeners as list of dictionaries
        :param listener_schema: schema to use to validate the listeners
        :param config: UserConfig object
        :param from_date: date from when to request notifications, if None it will be from now
        :param to_date: date until when to request notifications, if None it will be until now
        :return: number of listeners running
        """
        logger.debug("Calling listen in ListenerManager...")

        # first check the config
        if config is None:
            config = user_config.UserConfig()

        # Create the engine and listener factories
        engine_factory: ef.EngineFactory = ef.EngineFactory(config.notification_engine, Auth.get_auth(config))
        listener_factory: elf.EventListenerFactory = elf.EventListenerFactory(engine_factory, listener_schema)

        # read the payload key from the schema
        payload_key = listener_schema.get("payload")

        # Parse notification listeners
        event_listeners: List[EventListener] = []
        for ls in listeners:
            logger.debug(f"Reading listeners {ls}")
            try:
                for ev_listener in listener_factory.create_listeners(ls, from_date, to_date, payload_key):
                    event_listeners.append(ev_listener)
                logger.debug("Listener dictionary correctly parsed")
            except Exception as e:
                raise EventListenerException(f"Not able to load listener dictionary {ls}: {e}")

        # Add the listeners to the manager and run them
        logger.debug("Starting listeners...")
        self._add_listeners(event_listeners)
        if not self._run_listeners():
            if len(self.listeners) == 0:
                raise EventListenerException("Listeners could not start, please check logs")
            else:
                logger.error("One or more listeners were not able to start")

        # return the number of listeners running
        return len(self.listeners)



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import importlib
import json
import os
import re
from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum
from typing import Dict

TEMPLATE = r"\${[\w|\.]+}"
JSON_FOLDER = "/tmp/aviso"


class TriggerType(Enum):
    """
    Enum for the various triggers accepted by the system
    """

    log = ("log_trigger", "LogTrigger")
    function = ("function_trigger", "FunctionTrigger")
    command = ("command_trigger", "CommandTrigger")
    echo = ("echo_trigger", "EchoTrigger")
    post = ("post_trigger", "PostTrigger")

    def get_class(self):
        module = importlib.import_module("pyaviso.triggers." + self.value[0])
        return getattr(module, self.value[1])


class Trigger(ABC):
    """
    This class is an abstract class providing:
        - a common abstraction for the various type of triggers
    """

    def __init__(self, notification: Dict[str, any], params: Dict[str, any]):
        """
        :param notification: dictionary containing the attributes characterising a notification
        :param params: dictionary containing the attributes characterising the trigger as defined in the listener
        """
        self._params = params
        self._notification = notification

    @property
    def notification(self) -> Dict[str, any]:
        return self._notification

    @property
    def params(self) -> Dict[str, any]:
        return self._params

    @abstractmethod
    def execute(self):
        """
        Abstract method called by the thread in the run() method. This forces any child class to implement
        what is required for the execution of the specific trigger through a common interface.
        """
        pass

    def replace_template(self, text: str) -> str:
        """
        This method scans the text as input looking for the template pattern and replace it each match with the relative
        parameter taken from the notification dictionary
        :param text:
        :return:
        """
        matches = re.findall(TEMPLATE, text)
        for match in matches:
            assert len(match) > 3, "Wrong format for the variable templating, variable name must be specified"
            variable = match[2 : (len(match) - 1)]
            sub_pattern = f"\\{match}"
            if variable == "json":  # special case where we dump the whole notification dictionary
                json_dump = f"'{json.dumps(self.notification)}'"
                text = re.sub(sub_pattern, json_dump, text)
            elif variable == "jsonpath":  # special case where we save the notification dictionary to a json file
                if not os.path.exists(JSON_FOLDER) and not os.path.isdir(JSON_FOLDER):
                    os.makedirs(JSON_FOLDER, exist_ok=True)  # create folder first
                dtime = datetime.now().__str__().replace(" ", "")
                file_name = f"{JSON_FOLDER}/{dtime}.json"
                with open(file_name, "w") as file:
                    file.write(json.dumps(self.notification))
                text = re.sub(sub_pattern, file_name, text)
            else:
                # the variable may contain namespaces inside our nested dictionary
                split_variable = variable.split(".")
                string_to_eval = "self.notification"
                for v in split_variable:
                    string_to_eval += f"['{v}']"
                text = re.sub(sub_pattern, eval(string_to_eval), text)

        return text



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import json
import os
from shutil import rmtree
from typing import Dict, List

from . import logger
from .authentication.auth import Auth
from .custom_exceptions import ServiceConfigException
from .engine import engine_factory as ef
from .user_config import UserConfig

KEY_PREFIX = "/ec/config/"
MAX_FILE_SIZE_HARD = 1048576  # 1MiB


class ServiceConfigManager:
    """
    This class manages the operations associated to the configuration server
    """

    def __init__(self, config: UserConfig):
        """

        :param config: UserConfig object
        """
        logger.debug("Creating the engine")
        # Create the engine factory
        engine_factory: ef.EngineFactory = ef.EngineFactory(config.configuration_engine, Auth.get_auth(config))
        # Create the engine to connect to the configuration server
        self._engine = engine_factory.create_engine()
        self._max_file_size = config.configuration_engine.max_file_size

    def push(self, service: str, directory: str, user_message: str, delete: bool) -> List[str]:
        """
        This method implements the push command
        :param service: service to push to
        :param directory: directory from where to read the files to push
        :param user_message: message associated to this push operation
        :param delete: if True the files not pushed will be deleted from the server folder
        :return: List of files pushed
        """
        logger.debug("Calling push...")

        # Find the KV pairs in the dir
        if not directory.endswith("/"):
            directory = directory + "/"
        logger.debug(f"Looking for file in the directory {directory}")
        service_key = self._build_service_key(service)
        max_file_size_soft = self._max_file_size * 1024  # convert KiB to B
        kvs: List[Dict[str, any]] = []
        pushed_files: List[str] = []
        for x in os.walk(directory):
            for fp in x[2]:  # any file
                # prepare the key with suffix and prefix
                local_path = x[0]
                # removing the directory part so to have just the relative path
                suffix = local_path[len(directory) :]
                key: str = os.path.join(service_key, suffix, fp)
                fp_path = os.path.join(local_path, fp)

                # check file size limits
                if os.path.getsize(fp_path) > MAX_FILE_SIZE_HARD:
                    logger.warning(f"File {fp} exceeds hard limit of max file size allowed of {MAX_FILE_SIZE_HARD}B ")
                    continue
                if os.path.getsize(fp_path) > max_file_size_soft:
                    logger.warning(
                        f"File {fp} exceeds the configured max file size allowed of {max_file_size_soft}B, "
                        f"try increasing the limit in the configuration"
                    )
                    continue
                # prepare the value, read the file as binary
                with open(fp_path, "rb") as f:
                    value: bytes = f.read()
                kv: Dict[str, bytes] = {"key": key, "value": value}
                kvs.append(kv)
                pushed_files.append(f.name)
                logger.debug("Pushing file")

        # check if we need to delete any file
        ks_delete = []
        if delete:
            old_kvs = self._engine.pull(service_key)
            new_keys = []
            for new_kv in kvs:
                new_keys.append(new_kv["key"])
            for old_kv in old_kvs:
                old_key = old_kv["key"]
                if old_key not in new_keys:  # we need to delete this key
                    ks_delete.append(old_key)
                    logger.debug(f"Selecting file to delete: {old_key}")

        # push all the pairs as one transaction to the service
        if len(kvs) > 0:
            base_key = self._build_service_key(service, root_only=True)
            logger.debug("Calling the engine push with status update")
            if self._engine.push_with_status(kvs, base_key=base_key, message=user_message, ks_delete=ks_delete):
                logger.debug(f"Push operation for service {service} successfully executed")
                return pushed_files
            else:
                raise ServiceConfigException(f"Push operation for service {service} has failed")
        else:
            logger.debug(f"No file found to push to service {service}")
            return pushed_files

    def pull(self, service: str, key_only=False) -> List[List[any]]:
        """
        This method implements the pull command
        :param service: service to pull from
        :param key_only: if True no values are returned
        :return: KV pairs
        """
        logger.debug("Calling pull...")
        # pull the service
        service_key = self._build_service_key(service)
        kvs = self._engine.pull(service_key, key_only)
        if len(kvs) != 0:
            for kv in kvs:
                file_name = kv["key"]
                logger.debug(f"File pulled: {file_name}")
        else:
            logger.debug(f"No files found for service {service}")

        return kvs

    def remove(self, service: str) -> List[str]:
        """
        This method implements the remove command
        :param service: service to remove
        :return: List of files deleted
        """
        logger.debug("Calling remove...")
        # delete the service including the status
        service_key = self._build_service_key(service, root_only=True)
        kvs = self._engine.delete(service_key, prefix=True)
        removed_files = []
        if len(kvs) == 0:
            logger.debug(f"No file found for service {service}")
        else:
            for kv in kvs:
                file_name = kv["key"]
                logger.debug(f"File deleted: {file_name}")
                removed_files.append(file_name)
            logger.debug(f"Remove operation for service {service} successfully executed")
        return removed_files

    def pull_and_save(self, service: str, directory: str, delete: bool) -> List[str]:
        """
        This method implements the pull command and save the files retrieved to disk
        :param service: service to pull from
        :param directory: directory to save the file to
        :param delete: if True the files not pulled will be deleted from the local folder
        :return: List of files pulled
        """
        logger.debug("Calling pull and save...")

        service_key = self._build_service_key(service)

        # pull the service
        kvs: List[List[any]] = self._engine.pull(service_key)
        # look at the result
        pulled_files = []
        pulled_files_tmp = []
        if len(kvs) > 0:

            # First check if we need to delete the existing folder
            if delete and os.path.exists(directory):
                try:
                    # try to delete it
                    rmtree(directory)
                    logger.debug(f"Existing directory {directory} successfully deleted.")
                except Exception as e:
                    logger.warning(f"Error in deleting the folder {directory}: {e}")
                    logger.debug("", exc_info=True)

            # save the files retrieved with .tmp suffix
            for kv in kvs:
                # extract the file path and create the directory structure
                relative_file_path = kv["key"][len(service_key) :]
                full_path = os.path.join(directory, relative_file_path)
                full_path_tmp = full_path + ".tmp"
                os.makedirs(os.path.dirname(full_path_tmp), exist_ok=True)
                with open(full_path_tmp, "wb") as f:
                    f.write(kv["value"])
                    logger.debug(f"File successfully saved: {full_path_tmp}")
                    pulled_files.append(full_path)
                    pulled_files_tmp.append(full_path_tmp)

            # unlink existing files
            for f in pulled_files:
                try:
                    os.unlink(f)
                    logger.debug(f"File successfully unlinked: {f}")
                except FileNotFoundError:  # ignore
                    pass

            # remove the .tmp suffix
            for f_tmp, f in zip(pulled_files_tmp, pulled_files):
                os.rename(f_tmp, f)
                logger.debug(f"File successfully renamed: {f}")

            logger.debug(f"Pull operation for service {service} successfully executed")

        else:
            logger.debug(f"No file found for service {service}")

        return pulled_files

    def status(self, service: str) -> Dict[str, str]:
        """
        This method retrieves the status of the service passed
        :param service: service to check
        :return: status as dictionary
        """
        logger.debug("Calling status...")
        # pull the service
        service_key = self._build_service_key(service, root_only=True)
        kvs = self._engine.pull(service_key, prefix=False)
        if len(kvs) == 0:
            logger.warning(f"No service {service} found")
            return {}
        elif len(kvs) > 1:
            logger.error(f"Error in retrieving the status of service {service}, more than one KV returned")
            return {}

        logger.debug("Reading status from retrieved result")
        status = kvs[0]["value"].decode()
        status = json.loads(status)
        # add version to it
        status["version"] = kvs[0]["version"]
        logger.debug(f"Status retrieved for service {service}: {status}")

        return status

    def revert(self, service: str) -> List[str]:
        """
        This method reverts the service defined to the previous version
        :param service: service to revert
        :return: List of files reverted
        """
        #  first get the current version and revision
        logger.debug("Calling revert...")
        # pull the service
        service_key = self._build_service_key(service, root_only=True)
        kvs = self._engine.pull(service_key, key_only=True)
        reverted_files = []
        if len(kvs) == 0:
            logger.debug(f"No file found for service {service}")
            return reverted_files

        revert_kvs = []
        # for each file
        for kv in kvs:
            if kv["version"] == 1:  # we cannot revert this file
                continue
            # first we need to find the revision associated with the previous version
            target_version = kv["version"] - 1
            mod_rev = kv["mod_rev"]
            create_rev = kv["create_rev"]
            key = kv["key"]
            while mod_rev > create_rev:
                # retrieve the previous revision
                new_rev = mod_rev - 1
                old_kvs = self._engine.pull(key, key_only=True, rev=new_rev, prefix=False)
                if len(old_kvs) != 1:
                    raise ServiceConfigException(f"Error in retrieving older versions of key {key}, revision {new_rev}")
                new_kv = old_kvs[0]
                # it may be still the same version
                if new_kv["version"] == target_version:
                    # retrieve now the file of the target version and save it for later
                    old_kvs = self._engine.pull(key, rev=new_rev, prefix=False)
                    key = old_kvs[0]["key"]
                    revert_kvs.append(old_kvs[0])
                    logger.debug(f"File {key} waiting to be reverted to version {target_version}")
                    break

        if len(revert_kvs) == 0:
            logger.debug(f"No revertible file found for service {service}")
        else:
            # push it back to the store as a new version
            if self._engine.push(revert_kvs):
                for kv in revert_kvs:
                    key = kv["key"]
                    version = kv["version"]
                    reverted_files.append(key)
                    logger.debug(f"File {key} successfully reverted to version {version}")
            else:
                raise ServiceConfigException(f"Service {service} could not be reverted to previous version")

            logger.debug(f"Revert operation for service {service} successfully executed")

        return reverted_files

    def _build_service_key(self, service: str, root_only=False) -> str:
        """
        :param service:
        :param root_only: if True will return the service key with no "/" to indicate only the root key
        :return: the key associated to the folder of the service on the key-value store
        """
        suffix = "/"
        if service.endswith("/") or root_only:
            suffix = ""
        service_key = KEY_PREFIX + service + suffix
        return service_key



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import os
import subprocess
from typing import Dict

from .. import logger
from ..custom_exceptions import TriggerException
from . import trigger
from .trigger import TriggerType


class CommandTrigger(trigger.Trigger):
    """
    This class implements the 'Shell' trigger by executing an external shell script defined by the user.
    This class expects the params to contain the path of the script to run.
    Moreover this classes passes all the arguments defined together with NOTIFICATION_KEY and NOTIFICATION_VALUE as
    local variables.
    """

    def __init__(self, notification: Dict[str, any], params: Dict[str, any]):
        trigger.Trigger.__init__(self, notification, params)
        assert params.get("command") is not None, "command is a mandatory field"
        self.command: str = params.get("command")
        self.trigger_type = TriggerType.command

    def execute(self):
        logger.info("Starting Command Trigger...'")

        # prepare the variables passed as local variables
        my_env = os.environ.copy()  # don't change the caller environment
        if "environment" in self.params:
            envs = self.params.get("environment")
            for k in envs.keys():
                my_env[k] = self.replace_template(envs[k])

        # prepare command
        final_command = self.replace_template(self.command)
        # add working dir
        if "working_dir" in self.params:
            final_command = "cd " + self.params.get("working_dir") + ";" + final_command

        # create an independent process for the command
        logger.debug(f"Calling command {final_command}...")
        out = subprocess.Popen(final_command, env=my_env, shell=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE)
        # log the results
        stdout, stderr = out.communicate()
        if stdout is not None and stdout.decode() != "":
            logger.info(stdout.decode())
        if stderr is not None and stderr.decode() != "":
            raise TriggerException(stderr.decode())

        logger.debug("Command Trigger completed")



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from ..user_config import UserConfig
from .auth import Auth


class NoneAuth(Auth):
    """
    This class disables any authentication
    """

    def __init__(self, config: UserConfig):
        super(NoneAuth, self).__init__()

    def header(self):
        return {}



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from typing import Dict, List

import grpc
from etcd3 import Etcd3Client, etcdrpc

from .. import logger
from ..authentication.auth import Auth
from ..authentication.etcd_auth import EtcdAuth
from ..custom_exceptions import EngineException, EngineHistoryNotAvailableError
from ..user_config import EngineConfig
from .etcd_engine import MAX_KV_RETURNED, EtcdEngine


class EtcdGrpcEngine(EtcdEngine):
    """
    This class is a specialisation of the Engine class, able to connect to a etcd3 server directly via the gRPC
    interface. This class is relying on the pythonEtcd3 python module for implementing the gRPC protocol.
    """

    def __init__(self, config: EngineConfig, auth: Auth):
        super(EtcdGrpcEngine, self).__init__(config, auth)
        self._initialise_server()
        self._listening_list = []
        # set base url
        self._base_url = f"http://{self._host}:{self._port}/v3/"

    def _initialise_server(self):
        if type(self.auth) == EtcdAuth:
            self._server = Etcd3Client(
                self.host, self.port, user=self.auth.username, password=self.auth.password, timeout=self.timeout
            )
        else:
            self._server = Etcd3Client(self.host, self.port, timeout=self.timeout)

    def pull(
        self,
        key: str,
        key_only: bool = False,
        rev: int = None,
        prefix: bool = True,
        min_rev: int = None,
        max_rev: int = None,
    ) -> List[Dict[str, any]]:
        """
        This method implements a query to the notification server for all the key-values associated to the key as input.
        This key by default is a prefix, it can therefore return a set of key-values
        :param key: input in the query
        :param key_only: if True no values are returned
        :param rev: revision to pull
        :param prefix: if true the function will retrieve all the KV pairs starting with the key passed
        :param min_rev: if provided it filters for only KV pairs with mod_revision >= to min_rev
        :param max_rev: if provided it filters for only KV pairs with mod_revision <= to max_rev
        :return: List of key-value pairs formatted as dictionary
        """
        logger.debug(f"Calling pull for {key}...")

        # determine the range_end
        if prefix:
            range_end = self._incr_last_byte(key)
        else:
            range_end = None

        # call the get range on the ETCD_GRPC sever, order them newest first
        range_request = self._server._build_get_range_request(
            key=key, range_end=range_end, sort_order="descend", sort_target="key", keys_only=key_only
        )

        range_request.limit = MAX_KV_RETURNED
        if rev:
            range_request.revision = rev
        if min_rev:
            range_request.min_mod_revision = min_rev
        if max_rev:
            range_request.max_mod_revision = max_rev
        # make the call
        logger.debug(f"Pull request: {range_request}")
        try_again = True
        while try_again:
            try:
                try_again = False
                range_result = self._server.kvstub.Range(
                    range_request,
                    self._server.timeout,
                    credentials=self._server.call_credentials,
                    metadata=self._server.metadata,
                )
            except grpc._channel._InactiveRpcError as e:
                if e._state.code.name == "UNAUTHENTICATED":
                    # it seems that sometimes the token expires, so re-init the server and try again
                    try_again = True
                    logger.debug(f"Error {e}, trying again", exc_info=True)
                    self._initialise_server()
                elif (
                    e._state.code.name == "OUT_OF_RANGE" and "required revision has been compacted" in e._state.details
                ):
                    raise EngineHistoryNotAvailableError()
                else:
                    raise EngineException(e)
        logger.debug(f"Query for {key} completed")

        # parse the result to return just key-value pairs
        new_kvs: List[Dict[str, bytes]] = []
        logger.debug("Building key-value list")
        for kv in range_result.kvs:
            new_kv = self._parse_raw_kv(kv, key_only)
            new_kvs.append(new_kv)
            logger.debug(f"Key: {new_kv['key']} pulled successfully")

        logger.debug(f"{len(new_kvs)} keys found")
        return new_kvs

    def delete(self, key: str, prefix: bool = True) -> List[Dict[str, bytes]]:
        """
        This method deletes all the keys associated to this key, the key is a prefix as default
        :param key: key prefix to delete
        :param prefix: if true the function will delete all the KV pairs starting with the key passed
        :return: kvs deleted
        """
        logger.debug(f"Calling delete for {key}...")
        # determine the range_end
        if prefix:
            range_end = self._incr_last_byte(key)
        else:
            range_end = None

        # call the delete range on the ETCD_GRPC sever
        try_again = False
        del_request = self._server._build_delete_request(key=key, range_end=range_end, prev_kv=True)

        # make the call
        logger.debug(f"Deleting key range associated to key {key}")
        try_again = True
        while try_again:
            try:
                try_again = False
                del_result = self._server.kvstub.DeleteRange(
                    del_request,
                    self._server.timeout,
                    credentials=self._server.call_credentials,
                    metadata=self._server.metadata,
                )
            except grpc._channel._InactiveRpcError as e:
                if e._state.code.name == "UNAUTHENTICATED":
                    # it seems that sometimes the token expires, so re-init the server and try again
                    try_again = True
                    logger.debug(f"Error {e}, trying again", exc_info=True)
                    self._initialise_server()
                else:
                    raise e
        logger.debug(f"Delete request for key {key} completed")

        # parse the result to return just key-value pairs of what has been deleted
        del_kvs: List[Dict[str, bytes]] = []
        if hasattr(del_result, "prev_kvs"):
            logger.debug("Building key-value list")
            for kv in del_result.prev_kvs:
                new_kv = self._parse_raw_kv(kv)
                del_kvs.append(new_kv)
                logger.debug(f"Key: {new_kv['key']} deleted successfully")

        return del_kvs

    def push(self, kvs: List[Dict[str, any]], ks_delete: List[str] = None, ttl: int = None) -> bool:
        """
        Method to submit a list of key-value pairs and delete a list of keys from the server as a single transaction
        :param kvs: List of KV pair
        :param ks_delete: List of keys to delete before the push of the new ones. Note that each key is read as a folder
        :param ttl: time to leave of the keys pushed, once expired the keys will be deleted
        :return: True if successful
        """
        logger.debug("Calling push...")
        logger.debug("Preparing the transaction statement")
        ops = []

        # check if we need to request a lease for the ttl
        if ttl:
            try:
                lease = self._lease(ttl)
            except EngineException:
                raise EngineException("Not able to push keys")

        # first delete the keys requested
        if ks_delete is not None and len(ks_delete) != 0:
            for kd in ks_delete:
                # every key is deleted with prefix=True
                range_end = self._incr_last_byte(kd)
                delete = self._server._build_delete_request(kd, range_end)
                request_op = etcdrpc.RequestOp(request_delete_range=delete)
                ops.append(request_op)

        # Prepare the transaction with a put operation for each KV pair
        for kv in kvs:
            k = kv["key"]
            v = kv["value"]
            put = self._server._build_put_request(k, v)
            if ttl:
                put.lease = lease
            request_op = etcdrpc.RequestOp(request_put=put)
            ops.append(request_op)

        transaction_request = etcdrpc.TxnRequest(success=ops)

        # commit transaction
        # logger.debug(f"Committing the transaction statement: {ops}")
        try_again = True
        while try_again:
            try:
                try_again = False
                txn_response = self._server.kvstub.Txn(
                    transaction_request,
                    self._server.timeout,
                    credentials=self._server.call_credentials,
                    metadata=self._server.metadata,
                )
            except grpc._channel._InactiveRpcError as e:
                if e._state.code.name == "UNAUTHENTICATED":
                    # it seems that sometimes the token expires, so re-init the server and try again
                    try_again = True
                    logger.debug(f"Error {e}, trying again", exc_info=True)
                    self._initialise_server()
                else:
                    raise e
        assert txn_response.succeeded, "Not able to execute the transaction"
        logger.debug("Transaction completed")
        # read the header
        if hasattr(txn_response, "header"):
            h = txn_response.header
            rev = int(h.revision)
            logger.debug(f"New server revision {rev}")

        return True

    def lock(self, lock_id: str):
        """
        Acquire a lock identified by the id
        :param lock_id: Lock id
        :return: Lock if acquired otherwise exception if not acquired by the time the timeout expires
        """
        logger.debug("Calling lock...")
        try_again = True
        while try_again:
            try:
                try_again = False
                lock = self._server.lock(lock_id)
                res = lock.acquire(timeout=10)
            except grpc._channel._InactiveRpcError as e:
                if e._state.code.name == "UNAUTHENTICATED":
                    # it seems that sometimes the token expires, so re-init the server and try again
                    try_again = True
                    logger.debug(f"Error {e}, trying again", exc_info=True)
                    self._initialise_server()
                else:
                    raise e
            except Exception as e:
                raise EngineException(f"Not able to acquire lock {id}, {e}")
        if res:
            logger.debug(f"Lock {id} acquired")
            return lock
        else:
            raise EngineException(f"Not able to acquire lock {id}, timeout expired")

    def unlock(self, lock: any):
        """
        Acquire a lock identified by the id
        :param lock: Lock to release
        :return: True once released
        """
        logger.debug("Calling unlock...")
        try_again = True
        while try_again:
            try:
                try_again = False
                res = lock.release()
            except grpc._channel._InactiveRpcError as e:
                if e._state.code.name == "UNAUTHENTICATED":
                    # it seems that sometimes the token expires, so re-init the server and try again
                    try_again = True
                    logger.debug(f"Error {e}, trying again", exc_info=True)
                    self._initialise_server()
                else:
                    raise e

        logger.debug("Lock released")
        return res

    def _latest_revision(self, key: str) -> int:
        """
        :return: latest revision of the notification server.
        """
        logger.debug("Querying notification server for latest revision")

        # we need just the header back from the server
        range_request = self._server._build_get_range_request(key=key, keys_only=True)

        # make the call
        try_again = True
        while try_again:
            try:
                try_again = False
                range_result = self._server.kvstub.Range(
                    range_request,
                    self._server.timeout,
                    credentials=self._server.call_credentials,
                    metadata=self._server.metadata,
                )
            except grpc._channel._InactiveRpcError as e:
                if e._state.code.name == "UNAUTHENTICATED":
                    # it seems that sometimes the token expires, so re-init the server and try again
                    try_again = True
                    logger.debug(f"Error {e}, trying again", exc_info=True)
                    self._initialise_server()
                else:
                    raise e
        logger.debug("Query for latest revision completed")

        # read the header
        if hasattr(range_result, "header"):
            h = range_result.header
            rev = int(h.revision)
        else:
            raise EngineException("Error in reading server revision. Response does not contain header")

        logger.debug(f"Latest revision {rev}")
        return rev

    def _lease(self, ttl) -> str:
        """
        This method requests a Lease for the TTL specified
        :param ttl: Lease TTL
        :return: lease id
        """
        logger.debug(f"Calling lease for ttl {ttl}...")

        # create request
        lease_grant_request = etcdrpc.LeaseGrantRequest(TTL=ttl, ID=None)

        # make the call
        try_again = True
        while try_again:
            try:
                try_again = False
                res = self._server.leasestub.LeaseGrant(
                    lease_grant_request,
                    self._server.timeout,
                    credentials=self._server.call_credentials,
                    metadata=self._server.metadata,
                )
            except grpc._channel._InactiveRpcError as e:
                if e._state.code.name == "UNAUTHENTICATED":
                    # it seems that sometimes the token expires, so re-init the server and try again
                    try_again = True
                    logger.debug(f"Error {e}, trying again", exc_info=True)
                    self._initialise_server()
                else:
                    raise e
        if res:
            logger.debug(f"Lease {res.ID} acquired")
            return res.ID
        else:
            raise EngineException("Not able to acquire lease")

    def _parse_raw_kv(self, kv, key_only: bool = False) -> Dict[str, any]:
        """
        Internal method to translate the kv pair coming from the etcd server into a dictionary that fits better this
        application
        :param kv: raw kv pair from the etcd server
        :param key_only:
        :return: translated kv pair as dictionary
        """
        new_kv = {}
        if not key_only:
            new_kv["value"] = kv.value  # leave it as binary
        new_kv["key"] = kv.key.decode()
        new_kv["version"] = int(kv.version)
        new_kv["create_rev"] = int(kv.create_revision)
        new_kv["mod_rev"] = int(kv.mod_revision)
        return new_kv



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from ..user_config import UserConfig
from .etcd_auth import EtcdAuth


class EcmwfAuth(EtcdAuth):
    """
    This class set the authentication credentials for the ECMWF SSO authentication
    """

    def __init__(self, config: UserConfig):
        super(EcmwfAuth, self).__init__(config)
        self._password = config.password
        self._username = config.username

    def header(self):
        header = {"Authorization": f"EmailKey {self.username}:{self.password}"}
        return header
